import unittest
from switcheo.neo.utils import open_wallet
from switcheo.authenticated_client import AuthenticatedClient


ac = AuthenticatedClient(blockchain='neo')
testnet_privatekey_hexstring = '70f642894bc73dc50013be6d1dbe198f43237eaf9458d193c0b416c5385d9717'
kp = open_wallet(testnet_privatekey_hexstring)


class TestAuthenticatedClient(unittest.TestCase):

    def test_deposit(self):
        deposited_dict = {'result': 'ok'}
        deposit_dict = ac.deposit(asset="SWTH", amount=0.000001, private_key=kp)
        deposit_dict.pop('transaction_hash')
        self.assertDictEqual(deposit_dict, deposited_dict)
        deposit_dict = ac.deposit(asset="GAS", amount=0.000001, private_key=kp)
        deposit_dict.pop('transaction_hash')
        self.assertDictEqual(deposit_dict, deposited_dict)

    def test_withdrawal(self):
        swth_withdrawn_dict = {
            'event_type': 'withdrawal',
            'amount': '-100',
            'asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
            'blockchain': 'neo',
            'reason_code': 9,
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'transaction_hash': None,
            'contract_hash': '58efbb3cca7f436a55b1a05c0f36788d2d9a032e',
            'approval_transaction_hash': None,
            'group_index': 0
        }
        swth_withdrawal_dict = ac.withdrawal(asset="SWTH", amount=0.000001, private_key=kp)
        swth_withdrawal_dict.pop('id')
        swth_withdrawal_dict.pop('status')
        swth_withdrawal_dict.pop('created_at')
        swth_withdrawal_dict.pop('updated_at')
        self.assertDictEqual(swth_withdrawal_dict, swth_withdrawn_dict)

        gas_withdrawn_dict = {
            'event_type': 'withdrawal',
            'amount': '-100',
            'asset_id': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
            'blockchain': 'neo',
            'reason_code': 9,
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'transaction_hash': None,
            'contract_hash': '58efbb3cca7f436a55b1a05c0f36788d2d9a032e',
            'approval_transaction_hash': None,
            'group_index': 0
        }
        gas_withdrawal_dict = ac.withdrawal(asset="GAS", amount=0.000001, private_key=kp)
        gas_withdrawal_dict.pop('id')
        gas_withdrawal_dict.pop('status')
        gas_withdrawal_dict.pop('created_at')
        gas_withdrawal_dict.pop('updated_at')
        self.assertDictEqual(gas_withdrawal_dict, gas_withdrawn_dict)

    def test_create_and_cancel_order(self):
        order = ac.order(pair="SWTH_NEO", side="buy",
                         price=0.00001, quantity=10000, private_key=kp,
                         use_native_token=True, order_type="limit")
        ac.cancel_order(order_id=order['id'], private_key=kp)
        testnet_scripthash = 'fea2b883725ef2d194c9060f606cd0a0468a2c59'
        cancelled = False
        for trade in ac.get_orders(address=testnet_scripthash):
            if trade['id'] == order['id'] and\
                    trade['status'] == 'processed' and\
                    trade['makes'][0]['status'] in ['cancelled', 'cancelling']:
                cancelled = True
        self.assertTrue(cancelled)

    def test_order_filter(self):
        # Test side filter
        with self.assertRaises(ValueError):
            ac.order(pair="SWTH_NEO", side="test",
                     price=0.0001, quantity=100, private_key=kp,
                     use_native_token=True, order_type="limit")
        # Test order_type filter
        with self.assertRaises(ValueError):
            ac.order(pair="SWTH_NEO", side="buy",
                     price=0.0001, quantity=100, private_key=kp,
                     use_native_token=True, order_type="test")



import unittest
import time
from switcheo.public_client import PublicClient


pc = PublicClient(blockchain='neo')
pc_eth = PublicClient(blockchain='eth', contract_version='V1')


class TestPublicClient(unittest.TestCase):

    def test_get_exchange_status(self):
        exchange_status_dict = {'status': 'ok'}
        self.assertDictEqual(pc.get_exchange_status(), exchange_status_dict)

    def test_get_exchange_time(self):
        exchange_time_dict = {'timestamp': 1533362081336}
        self.assertGreater(pc.get_exchange_time()['timestamp'], exchange_time_dict['timestamp'])

    def test_get_token_details(self):
        exchange_token_list = ['NEO', 'GAS', 'SWTH', 'ETH']
        exchange_token_request = pc.get_token_details()
        self.assertTrue(set(exchange_token_request.keys()).issuperset(set(exchange_token_list)))

    def test_get_candlesticks(self):
        candles_key_list = ['time', 'open', 'close', 'high', 'low', 'volume', 'quote_volume']
        candles_request = pc.get_candlesticks(pair="SWTH_NEO",
                                              start_time=round(time.time()) - 360000,
                                              end_time=round(time.time()),
                                              interval=60)
        for candle in candles_request:
            candles_request = list(candle.keys())
        self.assertTrue(set(candles_request).issubset(set(candles_key_list)))

    def test_get_last_24_hours(self):
        last_stats_list = ['pair', 'open', 'close', 'high', 'low', 'volume', 'quote_volume']
        last_24_hours_request = pc.get_last_24_hours()
        for pair in last_24_hours_request:
            last_24_hours_request = list(pair.keys())
        self.assertTrue(set(last_24_hours_request).issubset(set(last_stats_list)))

    def test_get_last_price(self):
        last_price_dict = {
            'SWTH': ['ETH', 'NEO']
        }
        last_price_request = pc.get_last_price()
        for pair in last_price_request:
            last_price_request[pair] = list(last_price_request[pair].keys())
        self.assertTrue(set(last_price_request.keys()).issuperset(set(last_price_dict.keys())))
        for pair in last_price_dict:
            self.assertTrue(set(last_price_request[pair]).issubset(set(last_price_dict[pair])))

    def test_get_offers(self):
        offers_list = [{
            'id': '023bff30-ca83-453c-90e9-95502b52f492',
            'offer_asset': 'SWTH',
            'want_asset': 'NEO',
            'available_amount': 9509259,
            'offer_amount': 9509259,
            'want_amount': 9509259,
            'address': '7f345d1a031c4099540dbbbc220d4e5640ab2b6f'}]
        offers_set_list = set(offers_list[0].keys())
        offered_list = pc.get_offers()
        self.assertTrue(set(offered_list[0].keys()).issubset(offers_set_list))
        offered_set_list = set()
        for offer in offered_list:
            for key in offer.keys():
                offered_set_list.add(key)
        self.assertTrue(offered_set_list.issubset(offers_set_list))

        offered_list = pc_eth.get_offers(pair="JRC_ETH")
        self.assertTrue(set(offered_list[0].keys()).issubset(offers_set_list))
        offered_set_list = set()
        for offer in offered_list:
            for key in offer.keys():
                offered_set_list.add(key)
        self.assertTrue(offered_set_list.issubset(offers_set_list))

    def test_get_trades(self):
        trades_key_list = ['id', 'fill_amount', 'take_amount', 'event_time', 'is_buy']
        trades_list = pc.get_trades(pair="SWTH_NEO",
                                    limit=1,
                                    start_time=int(round(time.time())) - 2419200,
                                    end_time=int(round(time.time())))
        trades_list = trades_list[0].keys()
        self.assertTrue(set(trades_list).issubset(set(trades_key_list)))
        with self.assertRaises(ValueError):
            pc.get_trades(pair="SWTH_NEO", limit=0)
        with self.assertRaises(ValueError):
            pc.get_trades(pair="SWTH_NEO", limit=1000000)

    def test_get_pairs(self):
        all_pairs = ['GAS_NEO', 'SWTH_NEO', 'ACAT_NEO', 'APH_NEO', 'AVA_NEO', 'COUP_NEO', 'CPX_NEO']
        neo_pairs = ['GAS_NEO', 'SWTH_NEO', 'ACAT_NEO', 'APH_NEO', 'ASA_NEO', 'AVA_NEO', 'COUP_NEO', 'CPX_NEO',
                     'DBC_NEO', 'EFX_NEO', 'FTWX_NEO', 'GALA_NEO', 'GUARD_NEO', 'LRN_NEO', 'LX_NEO', 'MCT_NEO',
                     'NKNO_NEO', 'NOS_NEO', 'NRVE_NEO', 'OBT_NEO', 'ONT_NEO', 'PHX_NEO', 'PKC_NEO', 'PROQ_NEO',
                     'QLC_NEO', 'RCPT_NEO', 'RHT_NEO', 'RPX_NEO', 'SDS_NEO', 'SENNO_NEO', 'SOUL_NEO', 'TKY_NEO',
                     'TMN_NEO', 'TNC_NEO', 'TOLL_NEO', 'ZPT_NEO', 'BOLTC_NEO', 'MCTP_NEO', 'NRVEP_NEO', 'RHTC_NEO']
        self.assertTrue(set(pc.get_pairs(show_inactive=True)).issuperset(set(all_pairs)))
        self.assertTrue(set(pc.get_pairs(base="NEO", show_inactive=True)).issuperset(set(neo_pairs)))

    def test_get_contracts(self):
        contracts_dict = {
            'NEO': {
                'V1': '0ec5712e0f7c63e4b0fea31029a28cea5e9d551f',
                'V1_5': 'c41d8b0c30252ce7e8b6d95e9ce13fdd68d2a5a8',
                'V2': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                'V3': '58efbb3cca7f436a55b1a05c0f36788d2d9a032e'
            },
            'ETH': {
                'V1': '0x4dcf0244742e72309666db20d367f6dd196e884e'
            },
            'QTUM': {
                'V1': '0x2b25406b0000c3661e9c88890690fd4b5c7b4234'
            },
            'EOS': {
                'V1': 'oboluswitch4'
            }
        }
        self.assertDictEqual(pc.get_contracts(), contracts_dict)

    def test_get_orders(self):
        orders_list = [{
            'id': 'ecb6ee9e-de8d-46d6-953b-afcc976be1ae',
            'blockchain': 'neo',
            'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'pair': 'SWTH_NEO',
            'side': 'buy',
            'price': '0.000001',
            'quantity': '1000000',
            'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
            'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
            'offer_amount': '6000000',
            'want_amount': '30000000000',
            'transfer_amount': '0',
            'priority_gas_amount': '0',
            'use_native_token': True,
            'native_fee_transfer_amount': 0,
            'deposit_txn': None,
            'created_at': '2018-08-08T18:39:13.864Z',
            'status': 'processed',
            'order_status': 'processed',
            'txn': None,
            'offer_asset_blockchain': 'neo',
            'want_asset_blockchain': 'neo',
            'broadcast_cutoff_at': '2019-05-04T15:53:04.809Z',
            'scheduled_cancellation_at': None,
            'counterpart_swap': None,
            'fills': [],
            'fill_groups': [],
            'makes': []
        }]
        switcheo_orders_list = orders_list.copy()
        orders_list = orders_list[0].keys()
        testnet_scripthash = 'fea2b883725ef2d194c9060f606cd0a0468a2c59'
        all_orders = pc.get_orders(address=testnet_scripthash)
        switcheo_orders = pc.get_orders(address=testnet_scripthash, pair="SWTH_NEO")
        self.assertGreaterEqual(len(all_orders), len(switcheo_orders))
        self.assertTrue(set(all_orders[0].keys()).issubset(set(orders_list)))
        switcheo_orders_list_set = set()
        switcheo_orders_set = set()
        for order in switcheo_orders_list:
            switcheo_orders_list_set.add(order['offer_asset_id'])
            switcheo_orders_list_set.add(order['want_asset_id'])
        for order in switcheo_orders:
            switcheo_orders_set.add(order['offer_asset_id'])
            switcheo_orders_set.add(order['want_asset_id'])
        self.assertTrue(switcheo_orders_set.issubset(switcheo_orders_list_set))

    def test_get_balance(self):
        balance_dict = {
            'confirming': {},
            'confirmed': {
                'GAS': '90000000.0',
                'SWTH': '73580203956.0',
                'NEO': '113073528.0'},
            'locked': {
                'GAS': '9509259.0',
                'NEO': '4000000.0'}}
        balance_dict_set = set(balance_dict.keys())
        first_address = ['ca7316f459db1d3b444f57fe1ab875b3a607c200']
        second_address = ['fea2b883725ef2d194c9060f606cd0a0468a2c59']
        all_addresses = ['ca7316f459db1d3b444f57fe1ab875b3a607c200', 'fea2b883725ef2d194c9060f606cd0a0468a2c59']
        contracts = pc.get_contracts()
        all_contracts = []
        for chain in contracts:
            for contract in contracts[chain]:
                if chain == 'NEO':
                    all_contracts.append(contracts[chain][contract])
        first_balance_dict = pc.get_balance(addresses=first_address, contracts=all_contracts)
        first_balance_dict_set = set(first_balance_dict.keys())
        second_balance_dict = pc.get_balance(addresses=second_address, contracts=all_contracts)
        second_balance_dict_set = set(second_balance_dict.keys())
        all_balance_dict = pc.get_balance(addresses=all_addresses, contracts=all_contracts)
        all_balance_dict_set = set(all_balance_dict.keys())
        self.assertTrue(first_balance_dict_set.issubset(balance_dict_set))
        self.assertTrue(second_balance_dict_set.issubset(balance_dict_set))
        self.assertTrue(all_balance_dict_set.issubset(balance_dict_set))

        sum_balance_dict = {'confirmed': {
                                'GAS': str(int(float(first_balance_dict['confirmed']['GAS'])) + int(
                                    float(second_balance_dict['confirmed']['GAS']))),
                                'NEO': str(int(float(first_balance_dict['confirmed']['NEO'])) + int(
                                    float(second_balance_dict['confirmed']['NEO']))),
                                'SWTH': str(int(float(first_balance_dict['confirmed']['SWTH'])) + int(
                                    float(second_balance_dict['confirmed']['SWTH']))),
                            }}
        self.assertDictEqual(all_balance_dict['confirmed'], sum_balance_dict['confirmed'])



import unittest
from switcheo.neo.utils import create_offer_hash, encode_message, to_neo_asset_amount, private_key_to_hex, open_wallet,\
    neo_get_scripthash_from_address, neo_get_address_from_scripthash, neo_get_scripthash_from_private_key,\
    neo_get_public_key_from_private_key, sign_message, sign_transaction, sign_txn_array
from neocore.KeyPair import KeyPair


testnet_privatekey = b'p\xf6B\x89K\xc7=\xc5\x00\x13\xbem\x1d\xbe\x19\x8fC#~\xaf\x94X\xd1\x93\xc0\xb4\x16\xc58]\x97\x17'
kp = KeyPair(priv_key=testnet_privatekey)
testnet_privatekey_hexstring = '70f642894bc73dc50013be6d1dbe198f43237eaf9458d193c0b416c5385d9717'
testnet_scripthash = 'fea2b883725ef2d194c9060f606cd0a0468a2c59'
testnet_scripthash_uint160 = neo_get_scripthash_from_private_key(private_key=testnet_privatekey)
testnet_address = 'APuP9GsSCPJKrexPe49afDV8CQYubZGWd8'
testnet_publickey = '303231353534363535356234326164643737343933636332316462356461396639376163646666343966346433653739633239666363303361303661356539373662'

message = 'This is a test.'
encoded_message = '010001f0112254686973206973206120746573742e220000'
json_message = {"name": "John Smith", "age": 27, "siblings": ["Jane", "Joe"]}
json_encoded_message = '010001f0387b22616765223a32372c226e616d65223a224a6f686e20536d697468222c227369626c696e6773223a5b224a616e65222c224a6f65225d7d0000'

transaction_dict = {'hash': '72b74c96b9174e9b9e1b216f7e8f21a6475e6541876a62614df7c1998c6e8376',
                    'sha256': '2109cbb5eea67a06f5dd8663e10fcd1128e28df5721a25d993e05fe2097c34f3',
                    'type': 209,
                    'version': 1,
                    'attributes': [{'usage': 32, 'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'}],
                    'inputs': [{'prevHash': 'f09b3b697c580d1730cd360da5e1f0beeae00827eb2f0055cbc85a5a4dadd8ea', 'prevIndex': 0},
                               {'prevHash': 'c858e4d2af1e1525fa974fb2b1678caca1f81a5056513f922789594939ff713d', 'prevIndex': 31}],
                    'outputs': [{'assetId': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7', 'scriptHash': 'e707714512577b42f9a011f8b870625429f93573', 'value': 1e-08}],
                    'scripts': [],
                    'script': '0800e1f505000000001432e125258b7db0a0dffde5bd03b2b859253538ab14592c8a46a0d06c600f06c994d1f25e7283b8a2fe53c1076465706f73697467823b63e7c70a795a7615a38d1ba67d9e54c195a1',
                    'gas': 0}

transaction_array = [{
                        'id': 'e30a7fdf-779c-4623-8f92-8a961450d843',
                        'offer_hash': None,
                        'available_amount': None,
                        'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                        'offer_amount': '2000000',
                        'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                        'want_amount': '10000000000',
                        'filled_amount': None,
                        'txn': {
                            'offerHash': 'b45ddfb97ade5e0363d9e707dac9ad1c530448db263e86494225a0025006f968',
                            'hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                            'sha256': 'f0b70640627947584a2976edeb055a124ae85594db76453532b893c05618e6ca',
                            'invoke': {
                                'scriptHash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                                'operation': 'makeOffer',
                                'args': [
                                    '592c8a46a0d06c600f06c994d1f25e7283b8a2fe',
                                    '9b7cffdaa674beae0f930ebe6085af9093e5fe56b34a5c220ccdcf6efc336fc5',
                                    2000000,
                                    '32e125258b7db0a0dffde5bd03b2b859253538ab',
                                    10000000000,
                                    '65333061376664662d373739632d343632332d386639322d386139363134353064383433'
                                ]
                            },
                            'type': 209,
                            'version': 1,
                            'attributes': [
                                {
                                    'usage': 32,
                                    'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                                }, {
                                    'usage': 129,
                                    'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                                }, {
                                    'usage': 144,
                                    'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                                }, {
                                    'usage': 2,
                                    'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                                }
                            ],
                            'inputs': [
                                {
                                    'prevHash': '0fcfd792a9d20a7795255d1d3d3927f5968b9953e80d16ffd222656edf8fedbc',
                                    'prevIndex': 0
                                }, {
                                    'prevHash': 'c858e4d2af1e1525fa974fb2b1678caca1f81a5056513f922789594939ff713d',
                                    'prevIndex': 35
                                }
                            ],
                            'outputs': [
                                {
                                    'assetId': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
                                    'scriptHash': 'e707714512577b42f9a011f8b870625429f93573',
                                    'value': 1e-08
                                }
                            ],
                            'scripts': [],
                            'script': '0800e1f505000000001432e125258b7db0a0dffde5bd03b2b859253538ab14592c8a46a0d06c600f06c994d1f25e7283b8a2fe53c1076465706f73697467823b63e7c70a795a7615a38d1ba67d9e54c195a1',
                            'gas': 0
                        },
                        'cancel_txn': None,
                        'price': '0.0002',
                        'status': 'pending',
                        'created_at': '2018-08-05T10:38:37.731Z',
                        'transaction_hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                        'trades': []
                    }, {
                        'id': '7dac087c-3709-48ea-83e1-83eadfc4cbe5',
                        'offer_hash': None,
                        'available_amount': None,
                        'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                        'offer_amount': '2000000',
                        'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                        'want_amount': '10000000000',
                        'filled_amount': None,
                        'txn': {
                            'offerHash': 'b45ddfb97ade5e0363d9e707dac9ad1c530448db263e86494225a0025006f968',
                            'hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                            'sha256': 'f0b70640627947584a2976edeb055a124ae85594db76453532b893c05618e6ca',
                            'invoke': {
                                'scriptHash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                                'operation': 'makeOffer',
                                'args': [
                                    '592c8a46a0d06c600f06c994d1f25e7283b8a2fe',
                                    '9b7cffdaa674beae0f930ebe6085af9093e5fe56b34a5c220ccdcf6efc336fc5',
                                    2000000,
                                    '32e125258b7db0a0dffde5bd03b2b859253538ab',
                                    10000000000,
                                    '65333061376664662d373739632d343632332d386639322d386139363134353064383433'
                                ]
                            },
                            'type': 209,
                            'version': 1,
                            'attributes': [
                                {
                                    'usage': 32,
                                    'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                                }
                            ],
                            'inputs': [
                                {
                                    'prevHash': '0fcfd792a9d20a7795255d1d3d3927f5968b9953e80d16ffd222656edf8fedbc',
                                    'prevIndex': 0
                                }, {
                                    'prevHash': 'c858e4d2af1e1525fa974fb2b1678caca1f81a5056513f922789594939ff713d',
                                    'prevIndex': 35
                                }
                            ],
                            'outputs': [
                                {
                                    'assetId': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
                                    'scriptHash': 'e707714512577b42f9a011f8b870625429f93573',
                                    'value': 1e-08
                                }
                            ],
                            'scripts': [],
                            'script': '0800e1f505000000001432e125258b7db0a0dffde5bd03b2b859253538ab14592c8a46a0d06c600f06c994d1f25e7283b8a2fe53c1076465706f73697467823b63e7c70a795a7615a38d1ba67d9e54c195a1',
                            'gas': 0
                        },
                        'cancel_txn': None,
                        'price': '0.0002',
                        'status': 'pending',
                        'created_at': '2018-08-05T10:38:37.731Z',
                        'transaction_hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                        'trades': []
                    }]


class TestNeoUtils(unittest.TestCase):

    def test_sign_message(self):
        signed_message = 'f60f6896a87d6d35e62668958e08fb5a7c26c1c381c5ed6ad5b1ad9f6f6e826b1ab7695ad4ddd337438da1cd802b5e19dc6861700f4bd87a58dfa002cde11a3a'
        self.assertEqual(sign_message(encoded_message=encoded_message,private_key_hex=testnet_privatekey_hexstring),
                         signed_message)

    def test_sign_transaction(self):
        signed_transaction = 'cb511f7acf269d22690cf656b2f868fc10d12baff2f398ad175ae7e5a1e599f02d63a52d13aa3f3a6b57eaa0231e9fc4f1ab7900c34240033232c5a9f6b8214b'
        self.assertEqual(sign_transaction(transaction=transaction_dict, private_key_hex=testnet_privatekey_hexstring),
                         signed_transaction)

    def test_sign_array(self):
        pass
#         signed_array = {'e30a7fdf-779c-4623-8f92-8a961450d843': 'b1b821d7aa3c3d388370eba8e910de5c3605fcae2d584b0e89e932658f6b335a6aac65c52928e6eebf85919464897b8966a5a4dbcfd92eb28a3ae88299533f2c', '7dac087c-3709-48ea-83e1-83eadfc4cbe5': 'b1b821d7aa3c3d388370eba8e910de5c3605fcae2d584b0e89e932658f6b335a6aac65c52928e6eebf85919464897b8966a5a4dbcfd92eb28a3ae88299533f2c'}
#         self.assertEqual(sign_txn_array(messages=transaction_array, private_key_hex=testnet_privatekey_hexstring)['e30a7fdf-779c-4623-8f92-8a961450d843'],
#                          signed_array['e30a7fdf-779c-4623-8f92-8a961450d843'])

    def test_encode_message(self):
        self.assertEqual(encode_message(message=message), encoded_message)
        self.assertEqual(encode_message(message=json_message), json_encoded_message)

    def test_to_neo_asset_amount(self):
        self.assertEqual(to_neo_asset_amount(10), '1000000000')
        self.assertEqual(to_neo_asset_amount(.01), '1000000')
        with self.assertRaises(ValueError):
            to_neo_asset_amount(0.0000000000001)
        with self.assertRaises(ValueError):
            to_neo_asset_amount(100000000)

    def test_private_key_to_hex(self):
        self.assertEqual(private_key_to_hex(key_pair=kp), testnet_privatekey_hexstring)

    def test_neo_get_scripthash_from_address(self):
        self.assertEqual(neo_get_scripthash_from_address(address=testnet_address), testnet_scripthash)

    def test_neo_get_address_from_scripthash(self):
        self.assertEqual(neo_get_address_from_scripthash(scripthash=testnet_scripthash), testnet_address)

    def test_neo_get_public_key_from_private_key(self):
        self.assertEqual(neo_get_public_key_from_private_key(private_key=testnet_privatekey).ToString(), testnet_publickey)

    def test_neo_get_scripthash_from_private_key(self):
        self.assertEqual(str(neo_get_scripthash_from_private_key(private_key=testnet_privatekey)), testnet_scripthash)

    def test_open_wallet(self):
        self.assertEqual(open_wallet(testnet_privatekey_hexstring).PublicKey, kp.PublicKey)
        self.assertEqual(open_wallet(testnet_privatekey_hexstring).PrivateKey, kp.PrivateKey)
        self.assertEqual(open_wallet(testnet_privatekey_hexstring).GetAddress(), kp.GetAddress())

    def test_create_offer_hash(self):
        self.assertEqual(create_offer_hash(neo_address='APuP9GsSCPJKrexPe49afDV8CQYubZGWd8',
                                           offer_asset_amt=6000000,
                                           offer_asset_hash='c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                                           want_asset_amt=30000000000,
                                           want_asset_hash='ab38352559b8b203bde5fddfa0b07d8b2525e132',
                                           txn_uuid='ecb6ee9e-de8d-46d6-953b-afcc976be1ae'),
                         '95a9502f11c62b85cf790b83104c89d3198a3b4dac6ba8a0e19090a8ee2207c7')



import unittest
from switcheo.utils import get_epoch_milliseconds, num2hexstring, num2varint, reverse_hex,\
    stringify_message, current_contract_hash, Request
from switcheo.public_client import PublicClient


r = Request(api_url='https://jsonplaceholder.typicode.com/', api_version='')
s = Request()


class TestSwitcheoUtils(unittest.TestCase):

    def test_get_epoch_milliseconds(self):
        self.assertGreaterEqual(get_epoch_milliseconds(), 13)

    def test_num2hexstring(self):
        self.assertEqual(num2hexstring(0), '00')
        self.assertEqual(num2hexstring(255), 'ff')
        self.assertEqual(num2hexstring(256, size=2, little_endian=True), '0001')
        self.assertEqual(num2hexstring(2222, size=2, little_endian=True), 'ae08')

    def test_num2varint(self):
        self.assertEqual(num2varint(0), '00')
        self.assertEqual(num2varint(252), 'fc')
        self.assertEqual(num2varint(253), 'fdfd00')
        self.assertEqual(num2varint(255), 'fdff00')
        self.assertEqual(num2varint(256), 'fd0001')
        self.assertEqual(num2varint(2222), 'fdae08')
        self.assertEqual(num2varint(111111), 'fe07b20100')
        self.assertEqual(num2varint(11111111111), 'ffc719469602000000')

    def test_reverse_hex(self):
        self.assertEqual(reverse_hex('ABCD'), 'CDAB')
        self.assertEqual(reverse_hex('0000000005f5e100'), '00e1f50500000000')

    def test_stringify_message(self):
        json_msg = {"name": "John Smith", "age": 27, "siblings": ["Jane", "Joe"]}
        stringify_msg = '{"age":27,"name":"John Smith","siblings":["Jane","Joe"]}'
        self.assertEqual(stringify_message(json_msg), stringify_msg)

    def test_current_contract_hash(self):
        pc = PublicClient()
        expected_current_contract_dict = {
            'NEO': '58efbb3cca7f436a55b1a05c0f36788d2d9a032e',
            'ETH': '0x4dcf0244742e72309666db20d367f6dd196e884e',
            'QTUM': '0x2b25406b0000c3661e9c88890690fd4b5c7b4234',
            'EOS': 'oboluswitch4'
        }
        self.assertDictEqual(current_contract_hash(pc.contracts), expected_current_contract_dict)

    def test_request_get(self):
        json_msg = {
            "userId": 1,
            "id": 1,
            "title": "sunt aut facere repellat provident occaecati excepturi optio reprehenderit",
            "body": "quia et suscipit\nsuscipit recusandae consequuntur expedita et cum\nreprehenderit molestiae ut ut quas totam\nnostrum rerum est autem sunt rem eveniet architecto"}
        self.assertDictEqual(r.get(path='/posts/1'), json_msg)

    def test_request_post(self):
        json_dict = {
            'title': 'foo',
            'body': 'bar',
            'userId': 1}
        json_msg = {
            'id': 101,
            'title': 'foo',
            'body': 'bar',
            'userId': 1}
        self.assertDictEqual(r.post(path='/posts', json_data=json_dict), json_msg)

    def test_request_status(self):
        self.assertDictEqual(s.status(), {'status': 'ok'})



# -*- coding:utf-8 -*-
"""
Description:
    Functions for signing data to send to the Switcheo API by using NEO Utilities to cryptographically sign
    data that gets authenticated on the NEO network.
Usage:
    from switcheo.neo.signatures import <function_name>
"""

from switcheo.utils import get_epoch_milliseconds
from switcheo.neo.utils import sign_message, sign_transaction, sign_txn_array, encode_message,\
    neo_get_scripthash_from_private_key, private_key_to_hex


def sign_create_cancellation(cancellation_params, key_pair):
    """
    Function to sign the parameters required to create a cancellation request from the Switcheo Exchange.
    Execution of this function is as follows::

        sign_create_cancellation(cancellation_params=signable_params, key_pair=key_pair)

    The expected return result for this function is as follows::

        {
            'order_id': 'aa647b95-d546-4d29-961e-bd62b18b07bf',
            'timestamp': 1542092600331,
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'signature': '475bc3ecd2310201a3b5357b52b1866aaf5a5618932500e43503ebb....'
        }

    :param cancellation_params: Dictionary with Order ID and timestamp to sign for creating the cancellation.
    :type cancellation_params: dict
    :param key_pair: The KeyPair for the wallet being used to sign deposit message.
    :type key_pair: KeyPair
    :return: Dictionary of signed message to send to the Switcheo API.
    """
    encoded_message = encode_message(cancellation_params)
    create_params = cancellation_params.copy()
    create_params['address'] = neo_get_scripthash_from_private_key(private_key=key_pair.PrivateKey).ToString()
    create_params['signature'] = sign_message(encoded_message=encoded_message,
                                              private_key_hex=private_key_to_hex(key_pair=key_pair))
    return create_params


def sign_execute_cancellation(cancellation_params, key_pair):
    """
    Function to sign the parameters required to execute a cancellation request on the Switcheo Exchange.
    Execution of this function is as follows::

        sign_execute_cancellation(cancellation_params=signable_params, key_pair=key_pair)

    The expected return result for this function is as follows::

        {
            'signature': '6a40d6c011b7517f8fd3f2d0de32dd486adfd1d424d06d56c80eb....'
        }

    :param cancellation_params: Parameters the Switcheo Exchange returns from the create cancellation.
    :type cancellation_params: dict
    :param key_pair: The KeyPair for the wallet being used to sign deposit message.
    :type key_pair: KeyPair
    :return: Dictionary of signed message to send to the Switcheo API.
    """
    signature = sign_transaction(transaction=cancellation_params['transaction'],
                                 private_key_hex=private_key_to_hex(key_pair=key_pair))
    return {'signature': signature}


def sign_create_deposit(deposit_params, key_pair):
    """
    Function to create a deposit request by generating a transaction request from the Switcheo API.
    Execution of this function is as follows::

        sign_create_deposit(deposit_details=create_deposit, key_pair=key_pair)

    The expected return result for this function is as follows::

        {
            'blockchain': 'neo',
            'asset_id': 'SWTH',
            'amount': '100',
            'timestamp': 1542091927575,
            'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'signature': '24ef6c63964988a2efe5fe67f04f46fdc2f1504fb5....'
        }

    :param deposit_params: The parameters generated by the create deposit function that now requires signature.
    :type deposit_params: dict
    :param key_pair: The KeyPair for the wallet being used to sign deposit message.
    :type key_pair: KeyPair
    :return: Dictionary response of signed deposit request that is ready to be executed on the NEO blockchain.
    """
    encoded_message = encode_message(deposit_params)
    create_params = deposit_params.copy()
    create_params['address'] = neo_get_scripthash_from_private_key(private_key=key_pair.PrivateKey).ToString()
    create_params['signature'] = sign_message(encoded_message=encoded_message,
                                              private_key_hex=private_key_to_hex(key_pair=key_pair))
    return create_params


def sign_execute_deposit(deposit_params, key_pair):
    """
    Function to execute the deposit request by signing the transaction generated by the create deposit function.
    Execution of this function is as follows::

        sign_execute_deposit(deposit_details=create_deposit, key_pair=key_pair)

    The expected return result for this function is as follows::

        {
            'signature': '3cc4a5cb7b7d50383e799add2ba35382b6f2f1b2e3b97802....'
        }

    :param deposit_params: The parameters generated by the create deposit function that now requires signature.
    :type deposit_params: dict
    :param key_pair: The KeyPair for the wallet being used to sign deposit message.
    :type key_pair: KeyPair
    :return: Dictionary with the result status of the deposit attempt.
    """
    signature = sign_transaction(transaction=deposit_params['transaction'],
                                 private_key_hex=private_key_to_hex(key_pair=key_pair))
    return {'signature': signature}


def sign_create_order(order_params, key_pair):
    """
    Function to sign the create order parameters and send to the Switcheo API.
    Execution of this function is as follows::

        sign_create_order(order_params=signable_params, key_pair=key_pair)

    The expected return result for this function is as follows::

        {
            'blockchain': 'neo',
            'pair': 'SWTH_NEO',
            'side': 'buy',
            'price': '0.00001000',
            'want_amount': '1000000000000',
            'use_native_tokens': True,
            'order_type': 'limit',
            'timestamp': 1542091535839,
            'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'signature': '88e93c14a7d3c2cf30dec012ad5cb69f5ff26fe2a....'
        }

    :param order_params: Parameters to create an order to be submitted to the Switcheo Order Book.
    :type order_params: dict
    :param key_pair: The NEO key pair to be used to sign messages for the NEO Blockchain.
    :type key_pair: KeyPair
    :return: Dictionary of signed message to send to the Switcheo API.
    """
    encoded_message = encode_message(order_params)
    create_params = order_params.copy()
    create_params['address'] = neo_get_scripthash_from_private_key(private_key=key_pair.PrivateKey).ToString()
    create_params['signature'] = sign_message(encoded_message=encoded_message,
                                              private_key_hex=private_key_to_hex(key_pair=key_pair))
    return create_params


def sign_execute_order(order_params, key_pair):
    """
    Function to execute the order request by signing the transaction generated from the create order function.
    Execution of this function is as follows::

        sign_execute_order(order_params=signable_params, key_pair=key_pair)

    The expected return result for this function is as follows::

        {
            'signatures': {
                'fill_groups': {},
                'fills': {},
                'makes': {
                    '952defd3-ad8a-4db3-bbd1-27d58ff6c7bd': '3f5aa331a731a808fe260502421cbb06ae3d5ea5ddfb1....'
                }
            }
        }

    :param order_params: The parameters generated by the create function that now require signing.
    :type order_params: dict
    :param key_pair: The NEO key pair to be used to sign messages for the NEO Blockchain.
    :type key_pair: KeyPair
    :return: Dictionary of the signed transaction to place an order on the Switcheo Order Book.
    """
    execute_params = {
        'signatures': {
            'fill_groups': {},
            'fills': sign_txn_array(messages=order_params['fills'],
                                    private_key_hex=private_key_to_hex(key_pair=key_pair)),
            'makes': sign_txn_array(messages=order_params['makes'],
                                    private_key_hex=private_key_to_hex(key_pair=key_pair))
        }
    }
    return execute_params


def sign_create_withdrawal(withdrawal_params, key_pair):
    """
    Function to create the withdrawal request by signing the parameters necessary for withdrawal.
    Execution of this function is as follows::

        sign_create_withdrawal(withdrawal_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'blockchain': 'neo',
            'asset_id': 'SWTH',
            'amount': '100',
            'timestamp': 1542090737236,
            'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'signature': 'f66d604c0a80940bf70ce9e13c0fd47bc79de....'
        }

    :param withdrawal_params: Dictionary specifications for withdrawal from the Switcheo Smart Contract.
    :type withdrawal_params: dict
    :param key_pair: The NEO key pair to be used to sign messages for the NEO Blockchain.
    :type key_pair: KeyPair
    :return: Dictionary of parameters to be sent to the Switcheo API
    """
    encoded_message = encode_message(withdrawal_params)
    create_params = withdrawal_params.copy()
    create_params['address'] = neo_get_scripthash_from_private_key(private_key=key_pair.PrivateKey).ToString()
    create_params['signature'] = sign_message(encoded_message=encoded_message,
                                              private_key_hex=private_key_to_hex(key_pair=key_pair))
    return create_params


def sign_execute_withdrawal(withdrawal_params, key_pair):
    """
    Function to execute the withdrawal request by signing the transaction generated from the create withdrawal function.
    Execution of this function is as follows::

        sign_execute_withdrawal(withdrawal_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'id': '3e1c0802-b44e-4681-a94d-29c1dec2f518',
            'timestamp': 1542090738192,
            'signature': 'e05a7b7bd30eb85959d75ea634cee06ad35d96502a763ae40....'
        }

    :param withdrawal_params: Parameters passed from the create withdrawal function to be signed and confirmed.
    :type withdrawal_params: dict
    :param key_pair: The NEO key pair to be used to sign messages for the NEO Blockchain.
    :type key_pair: KeyPair
    :return: Dictionary of parameters to be sent to the Switcheo API
    """
    withdrawal_id = withdrawal_params['id']
    signable_params = {
        'id': withdrawal_id,
        'timestamp': get_epoch_milliseconds()
    }
    encoded_message = encode_message(signable_params)
    execute_params = signable_params.copy()
    execute_params['signature'] = sign_message(encoded_message=encoded_message,
                                               private_key_hex=private_key_to_hex(key_pair=key_pair))
    return execute_params



# -*- coding:utf-8 -*-
"""
Description:
    Fixed8

Usage:
    from neocore.Fixed8 import Fixed8
"""

from neocore.Fixed8 import Fixed8
from switcheo.utils import reverse_hex


def num2fixed8(number, size=8):
    if size % 1 != 0:
        raise ValueError('Fixed8 size {} is not a whole number.'.format(size))
    return SwitcheoFixed8(float("{:.8f}".format(number))).toReverseHex()[:size*2]


class SwitcheoFixed8(Fixed8):

    def toHex(self):
        hexstring = hex(round(self.value * self.D))[2:]
        return "{:0>16s}".format(hexstring)

    def toReverseHex(self):
        return reverse_hex(self.toHex())



import unittest
from switcheo.Fixed8 import SwitcheoFixed8, num2fixed8


class TestFixed8(unittest.TestCase):

    def test_fixed8_inheritance(self):
        self.assertEqual(SwitcheoFixed8(2050000).ToString(), '0.0205')
        self.assertEqual(SwitcheoFixed8(100000000).ToInt(), 1)

    def test_to_hex(self):
        self.assertEqual(SwitcheoFixed8(205).toHex(), '00000004c5e52d00')
        self.assertEqual(SwitcheoFixed8(0.0205).toHex(), '00000000001f47d0')

    def test_to_reverse_hex(self):
        self.assertEqual(SwitcheoFixed8(205).toReverseHex(), '002de5c504000000')
        self.assertEqual(SwitcheoFixed8(0.0205).toReverseHex(), 'd0471f0000000000')

    def test_num2fixed8(self):
        self.assertEqual(num2fixed8(205), '002de5c504000000')
        self.assertEqual(num2fixed8(0.0205), 'd0471f0000000000')
        with self.assertRaises(ValueError):
            num2fixed8(205, size=1.1)



# -*- coding:utf-8 -*-
"""
Description:
    Functions for signing data to send to the Switcheo API by using Ethereum Utilities to cryptographically sign
    data that gets authenticated on the Ethereum network.
Usage:
    from switcheo.ethereum.signatures import <function_name>
"""

import binascii
from switcheo.utils import stringify_message
from switcheo.ethereum.utils import sign_txn_array
from eth_utils import to_checksum_address, to_normalized_address
from eth_account.messages import defunct_hash_message
from eth_account.account import Account
from web3 import Web3, HTTPProvider


def sign_create_cancellation(cancellation_params, private_key):
    """
    Function to sign the parameters required to create a cancellation request from the Switcheo Exchange.
    Execution of this function is as follows::

        sign_create_cancellation(cancellation_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'order_id': '3125550a-04f9-4475-808b-42b5f89d6693',
            'timestamp': 1542088842108,
            'address': '0x32c46323b51c977814e05ef5e258ee4da0e4c3c3',
            'signature': 'dac70ca711bcfbeefbdead2158ef8b15fab1a1....'
        }

    :param cancellation_params: Dictionary with Order ID and timestamp to sign for creating the cancellation.
    :type cancellation_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :return: Dictionary of signed message to send to the Switcheo API.
    """
    hash_message = defunct_hash_message(text=stringify_message(cancellation_params))
    hex_message = binascii.hexlify(hash_message).decode()
    signed_message = binascii.hexlify(Account.signHash(hex_message, private_key=private_key)['signature']).decode()
    create_params = cancellation_params.copy()
    create_params['address'] = to_normalized_address(Account.privateKeyToAccount(private_key=private_key).address)
    create_params['signature'] = signed_message
    return create_params


def sign_execute_cancellation(cancellation_params, private_key):
    """
    Function to sign the parameters required to execute a cancellation request on the Switcheo Exchange.
    Execution of this function is as follows::

        sign_execute_cancellation(cancellation_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'signature': '0x65986ed2cb631d4999ce8b9c895a43f....'
        }

    :param cancellation_params: Parameters the Switcheo Exchange returns from the create cancellation.
    :type cancellation_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :return: Dictionary of signed message to send to the Switcheo API.
    """
    cancellation_sha256 = cancellation_params['transaction']['sha256']
    signed_sha256 = binascii.hexlify(
        Account.signHash(cancellation_sha256, private_key=private_key)['signature']).decode()
    return {'signature': '0x' + signed_sha256}


def sign_create_deposit(deposit_params, private_key):
    """
    Function to sign the deposit parameters required by the Switcheo API.
    Execution of this function is as follows::

        sign_create_deposit(deposit_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'blockchain': 'eth',
            'asset_id': 'ETH',
            'amount': '10000000000000000',
            'timestamp': 1542089346249,
            'contract_hash': '0x607af5164d95bd293dbe2b994c7d8aef6bec03bf',
            'address': '0x32c46323b51c977814e05ef5e258ee4da0e4c3c3',
            'signature': 'd4b8491d6514bff28b9f2caa440f51a93f31d....'
        }

    :param deposit_params: Parameters needed to deposit to the Switcheo API and signed in this function.
    :type deposit_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :return: Dictionary of signed message to send to the Switcheo API.
    """
    hash_message = defunct_hash_message(text=stringify_message(deposit_params))
    hex_message = binascii.hexlify(hash_message).decode()
    signed_message = binascii.hexlify(Account.signHash(hex_message, private_key=private_key)['signature']).decode()
    create_params = deposit_params.copy()
    create_params['address'] = to_normalized_address(Account.privateKeyToAccount(private_key=private_key).address)
    create_params['signature'] = signed_message
    return create_params


def sign_execute_deposit(deposit_params, private_key, infura_url):
    """
    Function to execute the deposit request by signing the transaction generated from the create deposit function.
    Execution of this function is as follows::

        sign_execute_deposit(deposit_params=create_deposit, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'transaction_hash': '0xcf3ea5d1821544e1686fbcb1f49d423b9ea9f42772ff9ecdaf615616d780fa75'
        }

    :param deposit_params: The parameters generated by the create function that now requires a signature.
    :type deposit_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :param infura_url: The URL used to broadcast the deposit transaction to the Ethereum network.
    :type infura_url: str
    :return: Dictionary of the signed transaction to initiate the deposit of ETH via the Switcheo API.
    """
    create_deposit_upper = deposit_params.copy()
    create_deposit_upper['transaction']['from'] = to_checksum_address(create_deposit_upper['transaction']['from'])
    create_deposit_upper['transaction']['to'] = to_checksum_address(create_deposit_upper['transaction']['to'])
    create_deposit_upper['transaction'].pop('sha256')
    signed_create_txn = Account.signTransaction(create_deposit_upper['transaction'], private_key=private_key)
    execute_signed_txn = binascii.hexlify(signed_create_txn['hash']).decode()

    # Broadcast transaction to Ethereum Network.
    Web3(HTTPProvider(infura_url)).eth.sendRawTransaction(signed_create_txn.rawTransaction)

    return {'transaction_hash': '0x' + execute_signed_txn}


def sign_create_order(order_params, private_key):
    """
    Function to sign the create order parameters and send to the Switcheo API.
    Execution of this function is as follows::

        sign_create_order(order_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'blockchain': 'eth',
            'pair': 'JRC_ETH',
            'side': 'buy',
            'price': '0.00000003',
            'want_amount': '3350000000000000000000000',
            'use_native_tokens': False,
            'order_type': 'limit',
            'timestamp': 1542089785915,
            'contract_hash': '0x607af5164d95bd293dbe2b994c7d8aef6bec03bf',
            'signature': '536306a2f2aee499ffd6584027029ee585293b3686....',
            'address': '0x32c46323b51c977814e05ef5e258ee4da0e4c3c3'
        }

    :param order_params: Parameters to create an order to be submitted to the Switcheo Order Book.
    :type order_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :return: Dictionary of signed message to send to the Switcheo API.
    """
    hash_message = defunct_hash_message(text=stringify_message(order_params))
    hex_message = binascii.hexlify(hash_message).decode()
    create_params = order_params.copy()
    signed_message = binascii.hexlify(Account.signHash(hex_message, private_key=private_key)['signature']).decode()
    create_params['signature'] = signed_message
    create_params['address'] = to_normalized_address(Account.privateKeyToAccount(private_key=private_key).address)
    return create_params


def sign_execute_order(order_params, private_key):
    """
    Function to execute the order request by signing the transaction generated from the create order function.
    Execution of this function is as follows::

        sign_execute_order(order_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'signatures': {
                'fill_groups': {},
                'fills': {},
                'makes': {
                    '392cd607-27ed-4702-880d-eab8d67a4201': '0x5f62c585e0978454cc89aa3b86d3ea6afbd80fc521....'
                }
            }
        }

    :param order_params: The parameters generated by the create function that now require signing.
    :type order_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :return: Dictionary of the signed transaction to place an order on the Switcheo Order Book.
    """
    execute_params = {
        'signatures': {
            'fill_groups': sign_txn_array(messages=order_params['fill_groups'], private_key=private_key),
            'fills': {},
            'makes': sign_txn_array(messages=order_params['makes'], private_key=private_key),
        }
    }
    return execute_params


def sign_create_withdrawal(withdrawal_params, private_key):
    """
    Function to create a withdrawal from the Switcheo Smart Contract.
    Execution of this function is as follows::

        sign_create_withdrawal(withdrawal_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'blockchain': 'eth',
            'asset_id': 'ETH',
            'amount': '10000000000000000',
            'timestamp': 1542090476102,
            'contract_hash': '0x607af5164d95bd293dbe2b994c7d8aef6bec03bf',
            'address': '0x32c46323b51c977814e05ef5e258ee4da0e4c3c3',
            'signature': '375ddce62e5b3676d5e94ebb9f9a8af5963b....'
        }

    :param withdrawal_params: The parameters to be signed and create a withdraw from Switcheo.
    :type withdrawal_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :return: Dictionary of the signed transaction to initiate the withdrawal of ETH via the Switcheo API.
    """
    hash_message = defunct_hash_message(text=stringify_message(withdrawal_params))
    hex_message = binascii.hexlify(hash_message).decode()
    signed_message = binascii.hexlify(Account.signHash(hex_message, private_key=private_key)['signature']).decode()
    create_params = withdrawal_params.copy()
    create_params['address'] = to_normalized_address(Account.privateKeyToAccount(private_key=private_key).address)
    create_params['signature'] = signed_message
    return create_params


def sign_execute_withdrawal(withdrawal_params, private_key):
    """
    Function to execute the withdrawal request by signing the transaction generated from the create withdrawal function.
    Execution of this function is as follows::

        sign_execute_withdrawal(withdrawal_params=signable_params, private_key=eth_private_key)

    The expected return result for this function is as follows::

        {
            'signature': '0x33656f88b364d344e5b04f6aead01cdd3ac084489c39a9efe88c9873249bf1954525b1....'
        }

    :param withdrawal_params: The parameters generated by the create function that now require signing.
    :type withdrawal_params: dict
    :param private_key: The Ethereum private key to sign the deposit parameters.
    :type private_key: str
    :return: Dictionary of the signed transaction hash and initiate the withdrawal of ETH via the Switcheo API.
    """
    withdrawal_sha256 = withdrawal_params['transaction']['sha256']
    signed_sha256 = binascii.hexlify(Account.signHash(withdrawal_sha256, private_key=private_key)['signature']).decode()
    return {'signature': '0x' + signed_sha256}



#
# switcheo/neo/utils.py
# Keith Smith
#
# For testnet requests to the Switcheo exchange

import math
import binascii
import base58
from neocore.Cryptography.Crypto import Crypto
from neocore.KeyPair import KeyPair
from neocore.Cryptography.Helper import scripthash_to_address
from switcheo.utils import num2hexstring, stringify_message, reverse_hex
from switcheo.neo.transactions import serialize_transaction


def sign_message(encoded_message, private_key_hex):
    return Crypto.Sign(message=encoded_message.strip(), private_key=private_key_hex).hex()


def sign_transaction(transaction, private_key_hex):
    serialized_transaction = serialize_transaction(transaction=transaction, signed=False)
    return sign_message(encoded_message=serialized_transaction, private_key_hex=private_key_hex)


def sign_txn_array(messages, private_key_hex):
    message_dict = {}
    for message in messages:
        message_dict[message['id']] = sign_transaction(transaction=message['txn'],
                                                       private_key_hex=private_key_hex)
    return message_dict


def encode_message(message):
    message_hex = binascii.hexlify(stringify_message(message).encode('utf-8')).decode()
    message_hex_length = hex(int(len(message_hex) / 2))[2:]
    return '010001f0' + message_hex_length + message_hex + '0000'


def to_neo_asset_amount(amount, power=8):
    if 0.00000001 < amount < 1000000:
        return "{:.0f}".format(amount * math.pow(10, power))
    else:
        raise ValueError('Asset amount {} outside of acceptable range {}-{}.'.format(amount, 0.00000001, 1000000))


def private_key_to_hex(key_pair):
    return bytes(key_pair.PrivateKey).hex()


def neo_get_scripthash_from_address(address):
    """
    Convert a Public Address String to a ScriptHash (Address) String.

    :param address: The Public address to convert.
    :type address: str
    :return: String containing the converted ScriptHash.
    """
    hash_bytes = binascii.hexlify(base58.b58decode_check(address))
    return reverse_hex(hash_bytes[2:].decode('utf-8'))


def neo_get_address_from_scripthash(scripthash):
    """
    Core methods for manipulating keys
    NEP2 <=> WIF <=> Private => Public => ScriptHash <=> Address
    Keys are arranged in order of derivation.
    Arrows determine the direction.
    """
    scripthash_bytes = binascii.unhexlify(reverse_hex(scripthash))
    return scripthash_to_address(scripthash_bytes)


def neo_get_public_key_from_private_key(private_key):
    kp = KeyPair(priv_key=private_key)
    return kp.PublicKey


def neo_get_scripthash_from_private_key(private_key):
    script = b'21' + neo_get_public_key_from_private_key(private_key).encode_point(True) + b'ac'
    return Crypto.ToScriptHash(data=script)


def open_wallet(private_key):
    pk = bytes.fromhex(private_key)
    return KeyPair(priv_key=pk)


def create_offer_hash(neo_address, offer_asset_hash, offer_asset_amt, want_asset_hash, want_asset_amt, txn_uuid):
    reverse_user_hash = reverse_hex(neo_get_scripthash_from_address(neo_address))
    reverse_offer_asset_hash = reverse_hex(offer_asset_hash)
    reverse_offer_amount = num2hexstring(number=offer_asset_amt, size=8, little_endian=True)
    reverse_want_asset_hash = reverse_hex(want_asset_hash)
    reverse_want_amount = num2hexstring(number=want_asset_amt, size=8, little_endian=True)
    nonce_hex = txn_uuid.encode('utf-8').hex()
    offer_key_bytes = reverse_user_hash + reverse_offer_asset_hash + reverse_want_asset_hash + reverse_offer_amount +\
                      reverse_want_amount + nonce_hex
    offer_hash = reverse_hex(Crypto.Hash256(binascii.a2b_hex(offer_key_bytes)).hex())
    return offer_hash



import unittest
from switcheo.neo.transactions import serialize_transaction, serialize_transaction_attribute,\
    serialize_transaction_input, serialize_transaction_output, serialize_witness, serialize_claim_exclusive,\
    serialize_contract_exclusive, serialize_invocation_exclusive


transaction_dict = {'hash': '72b74c96b9174e9b9e1b216f7e8f21a6475e6541876a62614df7c1998c6e8376',
                    'sha256': '2109cbb5eea67a06f5dd8663e10fcd1128e28df5721a25d993e05fe2097c34f3',
                    'type': 209,
                    'version': 1,
                    'attributes': [{'usage': 32, 'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'},
                                   {'usage': 32, 'data': '6a3d9b359fc17d711017daa6c0e14d6172a791ed'}],
                    'inputs': [{'prevHash': 'f09b3b697c580d1730cd360da5e1f0beeae00827eb2f0055cbc85a5a4dadd8ea', 'prevIndex': 0},
                               {'prevHash': 'c858e4d2af1e1525fa974fb2b1678caca1f81a5056513f922789594939ff713d', 'prevIndex': 31}],
                    'outputs': [{'assetId': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7', 'scriptHash': 'e707714512577b42f9a011f8b870625429f93573', 'value': 1e-08}],
                    'scripts': [],
                    'script': '0800e1f505000000001432e125258b7db0a0dffde5bd03b2b859253538ab14592c8a46a0d06c600f06c994d1f25e7283b8a2fe53c1076465706f73697467823b63e7c70a795a7615a38d1ba67d9e54c195a1',
                    'gas': 0}


class TestTransactions(unittest.TestCase):

    def test_serialize_transaction(self):
        serialized_transaction = 'd101520800e1f505000000001432e125258b7db0a0dffde5bd03b2b859253538ab14592c8a46a0d06c600f06c994d1f25e7283b8a2fe53c1076465706f73697467823b63e7c70a795a7615a38d1ba67d9e54c195a100000000000000000220592c8a46a0d06c600f06c994d1f25e7283b8a2fe206a3d9b359fc17d711017daa6c0e14d6172a791ed02ead8ad4d5a5ac8cb55002feb2708e0eabef0e1a50d36cd30170d587c693b9bf000003d71ff3949598927923f5156501af8a1ac8c67b1b24f97fa25151eafd2e458c81f0001e72d286979ee6cb1b7e65dfddfb2e384100b8d148e7758de42e4168b71792c6001000000000000007335f929546270b8f811a0f9427b5712457107e7'
        self.assertEqual(serialize_transaction(transaction=transaction_dict, signed=False), serialized_transaction)

    def test_serialize_transaction_attribute(self):
        serialized_attributes = []
        serialized_attribute_expected_list = ['20592c8a46a0d06c600f06c994d1f25e7283b8a2fe',
                                              '206a3d9b359fc17d711017daa6c0e14d6172a791ed']
        for attribute in transaction_dict['attributes']:
            serialized_attributes.append(serialize_transaction_attribute(attr=attribute))
        self.assertListEqual(serialized_attributes, serialized_attribute_expected_list)

    def test_serialize_transaction_input(self):
        serialized_inputs = []
        serialized_input_expected_list = ['ead8ad4d5a5ac8cb55002feb2708e0eabef0e1a50d36cd30170d587c693b9bf00000',
                                          '3d71ff3949598927923f5156501af8a1ac8c67b1b24f97fa25151eafd2e458c81f00']
        for txn_input in transaction_dict['inputs']:
            serialized_inputs.append(serialize_transaction_input(txn_input=txn_input))
        self.assertListEqual(serialized_inputs, serialized_input_expected_list)

    def test_serialize_transaction_output(self):
        serialized_outputs = []
        serialized_output_expected_list = ['e72d286979ee6cb1b7e65dfddfb2e384100b8d148e7758de42e4168b71792c6001000000000000007335f929546270b8f811a0f9427b5712457107e7']
        for txn_output in transaction_dict['outputs']:
            serialized_outputs.append(serialize_transaction_output(txn_output=txn_output))
        self.assertListEqual(serialized_outputs, serialized_output_expected_list)

    def test_serialize_witness(self):
        # This is not used by Switcheo and I can't find a good test transaction for this, will pass for now.
        pass

    def test_serialize_claim_exclusive(self):
        with self.assertRaises(ValueError):
            serialize_claim_exclusive(transaction=transaction_dict)
        # Switcheo will not be allowing for GAS claims so this should never be necessary.
        # transaction_claim_dict = transaction_dict.copy()
        # transaction_claim_dict['type'] = 2
        # self.assertEqual(serialize_claim_exclusive(transaction=transaction_claim_dict), '')

    def test_serialize_contract_exclusive(self):
        with self.assertRaises(ValueError):
            serialize_contract_exclusive(transaction=transaction_dict)
        transaction_contract_dict = transaction_dict.copy()
        transaction_contract_dict['type'] = 128
        self.assertEqual(serialize_contract_exclusive(transaction=transaction_contract_dict), '')

    def test_serialize_invocation_exclusive(self):
        serialized_invocation = '520800e1f505000000001432e125258b7db0a0dffde5bd03b2b859253538ab14592c8a46a0d06c600f06c994d1f25e7283b8a2fe53c1076465706f73697467823b63e7c70a795a7615a38d1ba67d9e54c195a10000000000000000'
        self.assertEqual(serialize_invocation_exclusive(transaction=transaction_dict), serialized_invocation)
        transaction_invocation_dict = transaction_dict.copy()
        transaction_invocation_dict['type'] = 128
        with self.assertRaises(ValueError):
            serialize_invocation_exclusive(transaction=transaction_invocation_dict)



import unittest
from switcheo.switcheo_client import SwitcheoClient


sc = SwitcheoClient()
testnet_address1 = 'APuP9GsSCPJKrexPe49afDV8CQYubZGWd8'
testnet_address2 = 'AFqt5vxyg4KKVTcTV4sR5oYMyUGCbrMQVt'


class TestSwitcheoClient(unittest.TestCase):

    def test_order_history(self):
        orders_list = [{
            'id': 'ecb6ee9e-de8d-46d6-953b-afcc976be1ae',
            'blockchain': 'neo',
            'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
            'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
            'pair': 'SWTH_NEO',
            'side': 'buy',
            'price': '0.00001',
            'quantity': '1000000000000',
            'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
            'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
            'offer_amount': '6000000',
            'want_amount': '30000000000',
            'transfer_amount': '0',
            'priority_gas_amount': '0',
            'use_native_token': True,
            'native_fee_transfer_amount': 0,
            'deposit_txn': None,
            'created_at': '2018-08-08T18:39:13.864Z',
            'status': 'processed',
            'order_status': 'cancelled',
            'txn': None,
            'offer_asset_blockchain': 'neo',
            'want_asset_blockchain': 'neo',
            'broadcast_cutoff_at': '2019-05-04T15:53:04.809Z',
            'scheduled_cancellation_at': None,
            'counterpart_swap': None,
            'fills': [],
            'fill_groups': [],
            'makes': []
            }]
        orders_list = orders_list[0].keys()
        all_orders = sc.order_history(address=testnet_address1)
        self.assertTrue(set(all_orders[0].keys()).issubset(set(orders_list)))

    def test_balance_current_contract(self):
        expected_balance_current_contract_key_set = set(['NEO', 'ETH', 'QTUM', 'EOS'])
        expected_balance_current_contract_child_key_set = set(['confirming', 'confirmed', 'locked'])
        balance_current_contract = sc.balance_current_contract(testnet_address1)
        balance_current_contract_key_set = set(balance_current_contract.keys())
        self.assertTrue(balance_current_contract_key_set.issubset(expected_balance_current_contract_key_set))
        balance_current_contract_child_key_set = set(balance_current_contract['NEO'].keys())
        self.assertTrue(
            balance_current_contract_child_key_set.issubset(expected_balance_current_contract_child_key_set))
        balance_current_contract_child_key_set = set(balance_current_contract['ETH'].keys())
        self.assertTrue(
            balance_current_contract_child_key_set.issubset(expected_balance_current_contract_child_key_set))

    def test_balance_by_contract(self):
        expected_balance_by_contract_key_set = set(['NEO', 'ETH', 'QTUM', 'EOS'])
        expected_balance_by_contract_child_key_set = set(['V1', 'V1_5', 'V2', 'V3'])
        expected_balance_by_contract_sub_key_set = set(['confirming', 'confirmed', 'locked'])
        balance_by_contract = sc.balance_by_contract(testnet_address1)
        balance_by_contract_key_set = set(balance_by_contract.keys())
        self.assertTrue(balance_by_contract_key_set.issubset(expected_balance_by_contract_key_set))
        balance_by_contract_neo = balance_by_contract['NEO']
        balance_by_contract_eth = balance_by_contract['ETH']
        balance_by_contract_child_key_set = set(balance_by_contract_neo.keys())
        self.assertTrue(balance_by_contract_child_key_set.issubset(expected_balance_by_contract_child_key_set))
        balance_by_contract_sub_key_set = set(balance_by_contract_neo['V1'])
        self.assertTrue(balance_by_contract_sub_key_set.issubset(expected_balance_by_contract_sub_key_set))
        balance_by_contract_child_key_set = set(balance_by_contract_eth.keys())
        self.assertTrue(balance_by_contract_child_key_set.issubset(expected_balance_by_contract_child_key_set))
        balance_by_contract_sub_key_set = set(balance_by_contract_eth['V1'])
        self.assertTrue(balance_by_contract_sub_key_set.issubset(expected_balance_by_contract_sub_key_set))

    def test_balance_by_address_by_contract(self):
        expected_balance_by_address_key_set = set([testnet_address1, testnet_address2])
        balance_by_address_key_set = set(sc.balance_by_address_by_contract(testnet_address1, testnet_address2).keys())
        self.assertTrue(balance_by_address_key_set.issubset(expected_balance_by_address_key_set))



import json
import time
from websocket import create_connection, WebSocketConnectionClosedException
from threading import Thread


class WebsocketsClient(object):
    def __init__(self,
                 url='wss://test-ws.switcheo.io',
                 products=None,
                 message_type="subscribe",
                 should_print=True,
                 auth=False,
                 api_key="",
                 api_secret="",
                 api_passphrase="",
                 channels=None):
        self.url = url
        self.products = products
        self.channels = channels
        self.type = message_type
        self.stop = True
        self.error = None
        self.ws = None
        self.thread = None
        self.keep_alive = None
        self.message_count = 0
        self.auth = auth
        self.api_key = api_key
        self.api_secret = api_secret
        self.api_passphrase = api_passphrase
        self.should_print = should_print

    def start(self):
        def _go():
            self._connect()
            self._listen()
            self._disconnect()

        self.stop = False
        self.on_open()
        self.thread = Thread(target=_go)
        self.keep_alive = Thread(target=self._keepalive)
        self.thread.start()

    def _connect(self):
        if self.products is None:
            self.products = ["SWTH_NEO"]
        elif not isinstance(self.products, list):
            self.products = [self.products]

        if self.url[-1] == "/":
            self.url = self.url[:-1]

        if self.channels is None:
            # sub_params = {'type': 'subscribe', 'product_ids': self.products}
            sub_params = {
                "blockchain": "neo",
                "contractHash": "91b83e96f2a7c4fdf0c1688441ec61986c7cae26",
                "pair": "SWTH_NEO"
            }
        else:
            # sub_params = {'type': 'subscribe', 'product_ids': self.products, 'channels': self.channels}
            sub_params = {
                "blockchain": "neo",
                "contractHash": "91b83e96f2a7c4fdf0c1688441ec61986c7cae26",
                "pair": "SWTH_NEO"
            }

        # if self.auth:
            # timestamp = str(time.time())
            # message = timestamp + 'GET' + '/users/self/verify'
            # auth_headers = get_auth_headers(timestamp, message, self.api_key, self.api_secret, self.api_passphrase)
            # sub_params['signature'] = auth_headers['CB-ACCESS-SIGN']
            # sub_params['key'] = auth_headers['CB-ACCESS-KEY']
            # sub_params['passphrase'] = auth_headers['CB-ACCESS-PASSPHRASE']
            # sub_params['timestamp'] = auth_headers['CB-ACCESS-TIMESTAMP']

        self.ws = create_connection(self.url)

        self.ws.send(json.dumps(sub_params))

    def _keepalive(self, interval=30):
        while self.ws.connected:
            self.ws.ping("keepalive")
            time.sleep(interval)

    def _listen(self):
        self.keep_alive.start()
        while not self.stop:
            try:
                data = self.ws.recv()
                msg = json.loads(data)
            except ValueError as e:
                self.on_error(e)
            except Exception as e:
                self.on_error(e)
            else:
                self.on_message(msg)

    def _disconnect(self):
        try:
            if self.ws:
                self.ws.close()
        except WebSocketConnectionClosedException as e:
            print(e)
            pass
        finally:
            self.keep_alive.join()

        self.on_close()

    def close(self):
        self.stop = True    # will only disconnect after next msg recv
        self._disconnect()  # force disconnect so threads can join
        self.thread.join()

    def on_open(self):
        if self.should_print:
            print("-- Subscribed! --\n")

    def on_close(self):
        if self.should_print:
            print("\n-- Socket Closed --")

    def on_message(self, msg):
        if self.should_print:
            print(msg)

    def on_error(self, e, data=None):
        self.error = e
        self.stop = True
        print('{} - data: {}'.format(e, data))


# if __name__ == "__main__":
#     import sys
#     import time
#
    #
    # class MyWebsocketClient(ws.WebsocketsClient):
    #     def on_open(self):
    #         self.url = 'wss://test-ws.switcheo.io'
    #         self.products = ['SWTH_NEO']
    #         self.message_count = 0
    #         print("Let's count the messages!")
    #
    #     def on_message(self, msg):
    #         print(json.dumps(msg, indent=4, sort_keys=True))
    #         self.message_count += 1
    #
    #     def on_close(self):
    #         print("-- Goodbye! --")

    # wsClient = MyWebsocketClient()
    # wsClient.start()
    # print(wsClient.url, wsClient.products)
    # try:
    #     while True:
    #         print("\nMessageCount =", "%i \n" % wsClient.message_count)
    #         time.sleep(1)
    # except KeyboardInterrupt:
    #     wsClient.close()

    # if wsClient.error:
    #     sys.exit(1)
    # else:
    #     sys.exit(0)


from socketIO_client_nexus import SocketIO, BaseNamespace, LoggingNamespace
import logging

logging.getLogger('requests').setLevel(logging.WARNING)
logging.basicConfig(level=logging.DEBUG)

class BooksNamespace(BaseNamespace):

   def on_event(self, event, *args):
       print('on_response', event, args)

   def on_connect(self):
       logging.info('connected')

   def on_disconnect(self):
       logging.info('disconnect')

class TradesNamespace(BaseNamespace):

   def on_event(self, event, *args):
       print('on_response', event, args)

   def on_connect(self):
       logging.info('connected')

   def on_disconnect(self):
       logging.info('disconnect')



url = 'https://ws.switcheo.io'
joinDict = {"blockchain":"neo","contractHash":"91b83e96f2a7c4fdf0c1688441ec61986c7cae26","pair":"SWTH_NEO"}


socketIO = SocketIO(url, None, LoggingNamespace, transports="websocket", verify=True )
#, cert=('client.crt', 'client.key'))

books  = socketIO.define(BooksNamespace, '/v2/books')
trades = socketIO.define(TradesNamespace, '/v2/books')
books.emit("join", joinDict)
trades.emit("join",joinDict)

socketIO.wait()



# -*- coding:utf-8 -*-
"""
Description:
    Public Client for the Switcheo decentralized exchange.
    It is not required to use your private key or WIF to interact with these endpoints which can be used
    to track the exchange state.
Usage:
    from switcheo.public_client import PublicClient
"""

from switcheo.utils import current_contract_hash, Request


class PublicClient(object):
    """
    This class allows the user to interact with the Switcheo decentralized exchange API to retrieve information about
    the available options on the exchange, including system health, time, trade offers, etc.
    """

    def __init__(self,
                 blockchain="neo",
                 contract_version='V3',
                 api_url='https://test-api.switcheo.network/',
                 api_version='/v2'):
        """

        :param blockchain: Choose which blockchain to trade on.  Allowed value are neo (future eth and qtum)
        :type blockchain: str
        :param api_url: The URL for the Switcheo API endpoint.
        :type api_url: str
        :param api_version: Choose the version of the Switcho API to use.
        :type api_version: str
        """
        self.request = Request(api_url=api_url, api_version=api_version, timeout=30)
        self.blockchain = blockchain
        self.blockchain_key = blockchain.upper()
        self.contracts = self.get_contracts()
        self.contract_version = contract_version.upper()
        self.contract_hash = self.contracts[self.blockchain_key][self.contract_version]
        self.current_contract_hash = current_contract_hash(self.contracts)[self.blockchain_key]

    def get_exchange_status(self):
        """
        Function to fetch the state of the exchange.
        Execution of this function is as follows::

            get_exchange_status()

        The expected return result for this function is as follows::

            {
                'status': 'ok'
            }

        :return: Dictionary in the form of a JSON message with the exchange status.
        """
        return self.request.status()

    def get_exchange_time(self):
        """
        Function to fetch the time on the exchange in to synchronized server time.
        Execution of this function is as follows::

            get_exchange_time()

        The expected return result for this function is as follows::

            {
                'timestamp': 1533362081336
            }

        :return: Dictionary in the form of a JSON message with the exchange epoch time in milliseconds.
        """
        return self.request.get(path='/exchange/timestamp')

    def get_exchange_message(self):
        """
        Function to fetch the Switcheo Exchange message to ensure pertinent information is handled gracefully.
        Execution of this function is as follows::

            get_exchange_message()

        The expected return result for this function is as follows::

            {
                'message': 'Welcome to Switcheo Beta.',
                'message_type': 'info'
            }

        :return: Dictionary containing the sum of all addresses smart contract balances by processing state.
        """
        return self.request.get(path='/exchange/announcement_message')

    def get_token_details(self, show_listing_details=False, show_inactive=False):
        """
        Function to fetch the available tokens available to trade on the Switcheo exchange.
        Execution of this function is as follows::

            get_token_details()
            get_token_details(show_listing_details=True)
            get_token_details(show_inactive=True)
            get_token_details(show_listing_details=True, show_inactive=True)

        The expected return result for this function is as follows::

            {
                'NEO': {
                    'hash': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                    'decimals': 8
                },
                'GAS': {
                    'hash': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
                    'decimals': 8
                },
                'SWTH': {
                    'hash': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                    'decimals': 8
                },
                ...
            }

        :param show_listing_details: Parameter flag to indicate whether or not to show the token listing details.
        :type show_listing_details: bool
        :param show_inactive: Flag to return the tokens that are no longer traded on the Switcheo Exchange.
        :type show_inactive: bool
        :return: Dictionary in the form of a JSON message with the available tokens for trade on the Switcheo exchange.
        """
        api_params = {
            "show_listing_details": show_listing_details,
            "show_inactive": show_inactive
        }
        return self.request.get(path='/exchange/tokens', params=api_params)

    def get_candlesticks(self, pair, start_time, end_time, interval):
        """
        Function to fetch trading metrics from the past 24 hours for all trading pairs offered on the exchange.
        Execution of this function is as follows::

            get_candlesticks(pair="SWTH_NEO",
                             start_time=round(time.time()) - 350000,
                             end_time=round(time.time()),
                             interval=360))

        The expected return result for this function is as follows::

            [{
                'time': '1533168000',
                'open': '0.00046835',
                'close': '0.00046835',
                'high': '0.00046835',
                'low': '0.00046835',
                'volume': '240315335.0',
                'quote_volume': '513110569018.0'
            },{
                'time': '1533081600',
                'open': '0.00046835',
                'close': '0.00046835',
                'high': '0.00046835',
                'low': '0.00046835',
                'volume': '1170875.0',
                'quote_volume': '2500000000.0'
            },
            ...
            ]

        :param pair: The trading pair used to request candle statistics.
        :type pair: str
        :param start_time: The start time (in epoch seconds) range for collecting candle statistics.
        :type start_time: int
        :param end_time: The end time (in epoch seconds) range for collecting candle statistics.
        :type end_time: int
        :param interval: The time interval (in minutes) for candle statistics.  Allowed values: 1, 5, 30, 60, 360, 1440
        :type interval: int
        :return: List of dictionaries containing the candles statistics based on the parameter filters.
        """
        api_params = {
            "pair": pair,
            "interval": interval,
            "start_time": start_time,
            "end_time": end_time,
            "contract_hash": self.contract_hash
        }
        return self.request.get(path='/tickers/candlesticks', params=api_params)

    def get_last_24_hours(self):
        """
        Function to fetch trading metrics from the past 24 hours for all trading pairs offered on the exchange.
        Execution of this function is as follows::

            get_last_24_hours()

        The expected return result for this function is as follows::

            [{
                'pair': 'SWTH_NEO',
                'open': '0.000407',
                'close': 0.00040911',
                'high': '0.00041492',
                'low': '0.00036',
                'volume': '34572662197.0',
                'quote_volume': '86879788270667.0'
            },{
                'pair': 'GAS_NEO',
                'open': '0.3',
                'close': '0.30391642',
                'high': '0.352',
                'low': '0.2925',
                'volume': '4403569005.0',
                'quote_volume': '14553283004.0'
            },{
            ....
            }]

        :return: List of dictionaries containing the statistics of each trading pair over the last 24 hours.
        """
        return self.request.get(path='/tickers/last_24_hours')

    def get_last_price(self, symbols=None, bases=None):
        """
        Function to fetch the most recently executed trade on the order book for each trading pair.
        Execution of this function is as follows::

            get_last_price()
            get_last_price(symbols=['SWTH','GAS'])
            get_last_price(bases=['NEO'])
            get_last_price(symbols=['SWTH','GAS'], bases=['NEO'])

        The expected return result for this function is as follows::

            {
                'SWTH': {
                    'GAS': '0.0015085',
                    'NEO': '0.00040911'
                },
                'GAS': {
                    'NEO': '0.30391642'
                },
                ....
            }

        :param symbols: The trading symbols to retrieve the last price on the Switcheo Exchange.
        :type symbols: list
        :param bases: The base pair to retrieve the last price of symbols on the Switcheo Exchange.
        :type bases: list
        :return: Dictionary of trade symbols with the most recently executed trade price.
        """
        api_params = {}
        if symbols is not None:
            api_params['symbols'] = symbols
        if bases is not None:
            api_params['bases'] = bases
        return self.request.get(path='/tickers/last_price', params=api_params)

    def get_offers(self, pair="SWTH_NEO"):
        """
        Function to fetch the open orders on the order book for the trade pair requested.
        Execution of this function is as follows::

            get_offers(pair="SWTH_NEO")

        The expected return result for this function is as follows::

            [{
                'id': '2716c0ca-59bb-4c86-8ee4-6b9528d0e5d2',
                'offer_asset': 'GAS',
                'want_asset': 'NEO',
                'available_amount': 9509259,
                'offer_amount': 30000000,
                'want_amount': 300000000,
                'address': '7f345d1a031c4099540dbbbc220d4e5640ab2b6f'
            }, {
                ....
            }]

        :param pair: The trading pair that will be used to request open offers on the order book.
        :type pair: str
        :return: List of dictionaries consisting of the open offers for the requested trading pair.
        """
        api_params = {
            "pair": pair,
            "contract_hash": self.contract_hash
        }
        return self.request.get(path='/offers', params=api_params)

    def get_offer_book(self, pair="SWTH_NEO"):
        """
        Function to fetch the open orders formatted on the order book for the trade pair requested.
        Execution of this function is as follows::

            get_offer_book(pair="SWTH_NEO")

        The expected return result for this function is as follows::

            {
                'asks': [{
                    'price': '0.00068499',
                    'quantity': '43326.8348443'
                }, {
                    'price': '0.000685',
                    'quantity': '59886.34'
                }, {
                    ....
                }],
                'bids': [{
                    'price': '0.00066602',
                    'quantity': '3255.99999999'
                }, {
                    'price': '0.00066601',
                    'quantity': '887.99999999'
                }, {
                    ....
                }]
            }

        :param pair: The trading pair that will be used to request open offers on the order book.
        :type pair: str
        :return: List of dictionaries consisting of the open offers for the requested trading pair.
        """
        api_params = {
            "pair": pair,
            "contract_hash": self.contract_hash
        }
        return self.request.get(path='/offers/book', params=api_params)

    def get_trades(self, pair="SWTH_NEO", start_time=None, end_time=None, limit=5000):
        """
        Function to fetch a list of filled trades for the parameters requested.
        Execution of this function is as follows::

            get_trades(pair="SWTH_NEO", limit=3)

        The expected return result for this function is as follows::

            [{
                'id': '15bb16e2-7a80-4de1-bb59-bcaff877dee0',
                'fill_amount': 100000000,
                'take_amount': 100000000,
                'event_time': '2018-08-04T15:00:12.634Z',
                'is_buy': True
            }, {
                'id': 'b6f9e530-60ff-46ff-9a71-362097a2025e',
                'fill_amount': 47833882,
                'take_amount': 97950000000,
                'event_time': '2018-08-03T02:44:47.706Z',
                'is_buy': True
            }, {
                'id': '7a308ccc-b7f5-46a3-bf6b-752ab076cc9f',
                'fill_amount': 1001117,
                'take_amount': 2050000000,
                'event_time': '2018-08-03T02:32:50.703Z',
                'is_buy': True
            }]

        :param pair: The trading pair that will be used to request filled trades.
        :type pair: str
        :param start_time: Only return trades after this time (in epoch seconds).
        :type start_time: int
        :param end_time: Only return trades before this time (in epoch seconds).
        :type end_time: int
        :param limit: The number of filled trades to return. Min: 1, Max: 10000, Default: 5000
        :type limit: int
        :return: List of dictionaries consisting of filled orders that meet requirements of the parameters passed to it.
        """
        if limit > 10000 or limit < 1:
            raise ValueError("Attempting to request more trades than allowed by the API.")
        api_params = {
            "blockchain": self.blockchain,
            "pair": pair,
            "contract_hash": self.contract_hash
        }
        if start_time is not None:
            api_params['from'] = start_time
        if end_time is not None:
            api_params['to'] = end_time
        if limit != 5000:
            api_params['limit'] = limit
        return self.request.get(path='/trades', params=api_params)

    def get_recent_trades(self, pair="SWTH_NEO"):
        """
        Function to fetch a list of the 20 most recently filled trades for the parameters requested.
        Execution of this function is as follows::

            get_recent_trades(pair="SWTH_NEO")

        The expected return result for this function is as follows::

            [{
                'id': '15bb16e2-7a80-4de1-bb59-bcaff877dee0',
                'fill_amount': 100000000,
                'take_amount': 100000000,
                'event_time': '2018-08-04T15:00:12.634Z',
                'is_buy': True
            }, {
                'id': 'b6f9e530-60ff-46ff-9a71-362097a2025e',
                'fill_amount': 47833882,
                'take_amount': 97950000000,
                'event_time': '2018-08-03T02:44:47.706Z',
                'is_buy': True
            }, ...., {
                'id': '7a308ccc-b7f5-46a3-bf6b-752ab076cc9f',
                'fill_amount': 1001117,
                'take_amount': 2050000000,
                'event_time': '2018-08-03T02:32:50.703Z',
                'is_buy': True
            }]

        :param pair: The trading pair that will be used to request filled trades.
        :type pair: str
        :return: List of 20 dictionaries consisting of filled orders for the trade pair.
        """
        api_params = {
            "pair": pair
        }
        return self.request.get(path='/trades/recent', params=api_params)

    def get_pairs(self, base=None, show_details=False, show_inactive=False):
        """
        Function to fetch a list of trading pairs offered on the Switcheo decentralized exchange.
        Execution of this function is as follows::

            get_pairs()                  # Fetch all pairs
            get_pairs(base="SWTH")       # Fetch only SWTH base pairs
            get_pairs(show_details=True) # Fetch all pairs with extended information !Attention return value changes!

        The expected return result for this function is as follows::

            [
                'GAS_NEO',
                'SWTH_NEO',
                'MCT_NEO',
                'NKN_NEO',
                ....
                'SWTH_GAS',
                'MCT_GAS',
                'NKN_GAS',
                ....
                'MCT_SWTH',
                'NKN_SWTH'
            ]

        If you use the show_details parameter the server return a list with dictionaries as follows::

             [
                {'name': 'GAS_NEO', 'precision': 3},
                {'name': 'SWTH_NEO', 'precision': 6},
                {'name': 'ACAT_NEO', 'precision': 8},
                {'name': 'APH_NEO', 'precision': 5},
                {'name': 'ASA_NEO', 'precision': 8},
                ....
            ]

        :param base: The base trade pair to optionally filter available trade pairs.
        :type base: str
        :param show_details: Extended information for the pairs.
        :type show_details: bool
        :return: List of trade pairs available for trade on Switcheo.
        """
        api_params = {}
        if show_details:
            api_params["show_details"] = show_details
        if show_inactive:
            api_params["show_inactive"] = show_inactive
        if base is not None and base in ["NEO", "GAS", "SWTH", "USD", "ETH"]:
            api_params["bases"] = [base]
        return self.request.get(path='/exchange/pairs', params=api_params)

    def get_contracts(self):
        """
        Function to fetch the contract hashes for each smart contract deployed on the defined blockchain.
        Execution of this function is as follows::

            get_contracts()

        The expected return result for this function is as follows::

            {
                'NEO': {
                    'V1': '0ec5712e0f7c63e4b0fea31029a28cea5e9d551f',
                    'V1_5': '01bafeeafe62e651efc3a530fde170cf2f7b09bd',
                    'V2': '91b83e96f2a7c4fdf0c1688441ec61986c7cae26'
                },
                'ETH': {
                    'V1': '0x0000000000000000000000000000000000000000'
                }
            }

        :return: Dictionary containing the list of smart contract hashes per blockchain and version.
        """
        return self.request.get(path='/contracts')

    def get_orders(self, address, chain_name='NEO', contract_version='V3', pair=None, from_epoch_time=None,
                   order_status=None, before_id=None, limit=50):
        """
        Function to fetch the order history of the given address.
        Execution of this function is as follows::

            get_orders(address=neo_get_scripthash_from_address(address=address))

        The expected return result for this function is as follows::

            [{
                'id': '7cbdf481-6acf-4bf3-a1ed-4773f31e6931',
                'blockchain': 'neo',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'side': 'buy',
                'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'offer_amount': '53718500',
                'want_amount': '110000000000',
                'transfer_amount': '0',
                'priority_gas_amount': '0',
                'use_native_token': True,
                'native_fee_transfer_amount': 0,
                'deposit_txn': None,
                'created_at': '2018-08-03T02:44:47.692Z',
                'status': 'processed',
                'fills': [{
                    'id': 'b6f9e530-60ff-46ff-9a71-362097a2025e',
                    'offer_hash': '95b3b03be0bff8f58aa86a8dd599700bbaeaffc05078329d5b726b6b995f4cda',
                    'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                    'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                    'fill_amount': '47833882',
                    'want_amount': '97950000000',
                    'filled_amount': '',
                    'fee_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                    'fee_amount': '73462500',
                    'price': '0.00048835',
                    'txn': None,
                    'status': 'success',
                    'created_at': '2018-08-03T02:44:47.706Z',
                    'transaction_hash': '694745a09e33845ec008cfb79c73986a556e619799ec73274f82b30d85bda13a'
                }],
                'makes': [{
                    'id': '357088a0-cc80-49ab-acdd-980589c2d7d8',
                    'offer_hash': '420cc85abf02feaceb1bcd91489a0c1949c972d2a9a05ae922fa15d79de80c00',
                    'available_amount': '0',
                    'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                    'offer_amount': '5884618',
                    'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                    'want_amount': '12050000000',
                    'filled_amount': '0.0',
                    'txn': None,
                    'cancel_txn': None,
                    'price': '0.000488350041493775933609958506224066390041494',
                    'status': 'cancelled',
                    'created_at': '2018-08-03T02:44:47.708Z',
                    'transaction_hash': '1afa946546550151bbbd19f197a87cec92e9be58c44ec431cae42076298548b7',
                    'trades': []
                }]
            }, {
            ....
            }]

        :param address: The ScriptHash of the address to filter orders for.
        :type address: str
        :param pair: The trading pair to filter order requests on.
        :type pair: str
        :param chain_name: The name of the chain to find orders against.
        :type chain_name: str
        :param contract_version: The version of the contract to find orders against.
        :type contract_version: str
        :param from_epoch_time: Only return orders that are last updated at or after this time.
        :type from_epoch_time: int
        :param order_status: Only return orders have this status. Possible values are open, cancelled, completed.
        :type order_status: str
        :param before_id: Only return orders that are created before the order with this id.
        :type before_id: str
        :param limit: Only return up to this number of orders (min: 1, max: 200, default: 50).
        :type limit: int
        :return: List of dictionaries containing the orders for the given NEO address and (optional) trading pair.
        """
        api_params = {
            "address": address,
            "contract_hash": self.get_contracts()[chain_name.upper()][contract_version.upper()],
            "limit": limit
        }
        if pair is not None:
            api_params['pair'] = pair
        if from_epoch_time is not None:
            api_params['from_epoch_time'] = from_epoch_time
        if order_status is not None:
            api_params['order_status'] = order_status
        if before_id is not None:
            api_params['before_id'] = before_id
        return self.request.get(path='/orders', params=api_params)

    def get_balance(self, addresses, contracts):
        """
        Function to fetch the current account balance for the given address in the Switcheo smart contract.
        Execution of this function is as follows::

            get_balance(address=neo_get_scripthash_from_address(address=address))

        The expected return result for this function is as follows::

            {
                'confirming': {},
                'confirmed': {
                    'GAS': '100000000.0',
                    'SWTH': '97976537500.0',
                    'NEO': '52166118.0'
                },
                'locked': {}
            }

        :param addresses: The ScriptHash of the address(es) to retrieve its Smart Contract balance.
        :type addresses: list
        :param contracts: The contract hash(es) to retrieve all addresses Smart Contract balance.
        :type contracts: list
        :return: Dictionary containing the sum of all addresses smart contract balances by processing state.
        """
        api_params = {
            "addresses[]": addresses,
            "contract_hashes[]": contracts
        }
        return self.request.get(path='/balances', params=api_params)



# -*- coding:utf-8 -*-
"""
Description:
    Authenticated Client for the Switcheo decentralized exchange.
    This API extends the PublicClient by offering API endpoints that require signatures from the
    users private key.
    Deposit, Withdrawal, Trading, and Cancellations on the exchange are accessible through this class.
Usage:
    from switcheo.authenticated_client import AuthenticatedClient
"""

from functools import partial
from switcheo.public_client import PublicClient
from switcheo.utils import get_epoch_milliseconds
from switcheo.neo.utils import to_neo_asset_amount
from switcheo.neo.signatures import sign_create_deposit as sign_create_deposit_neo,\
    sign_execute_deposit as sign_execute_deposit_neo, sign_create_order as sign_create_order_neo,\
    sign_execute_order as sign_execute_order_neo, sign_create_withdrawal as sign_create_withdrawal_neo,\
    sign_execute_withdrawal as sign_execute_withdrawal_neo, sign_create_cancellation as sign_create_cancellation_neo,\
    sign_execute_cancellation as sign_execute_cancellation_neo
from switcheo.ethereum.signatures import sign_create_deposit as sign_create_deposit_eth,\
    sign_execute_deposit as sign_execute_deposit_eth, sign_create_order as sign_create_order_eth,\
    sign_execute_order as sign_execute_order_eth, sign_create_withdrawal as sign_create_withdrawal_eth,\
    sign_execute_withdrawal as sign_execute_withdrawal_eth, sign_create_cancellation as sign_create_cancellation_eth,\
    sign_execute_cancellation as sign_execute_cancellation_eth
from eth_utils import to_wei


class AuthenticatedClient(PublicClient):

    def __init__(self,
                 blockchain='neo',
                 contract_version='V3',
                 api_url='https://test-api.switcheo.network/',
                 api_version='/v2'):
        PublicClient.__init__(self,
                              blockchain=blockchain,
                              contract_version=contract_version,
                              api_url=api_url,
                              api_version=api_version)
        self.infura_dict = {
            'https://api.switcheo.network': 'https://infura.io/',
            'https://api.switcheo.network/': 'https://infura.io/',
            'api.switcheo.network': 'https://infura.io/',
            'api.switcheo.network/': 'https://infura.io/',
            'https://test-api.switcheo.network': 'https://ropsten.infura.io/',
            'https://test-api.switcheo.network/': 'https://ropsten.infura.io/',
            'test-api.switcheo.network': 'https://ropsten.infura.io/',
            'test-api.switcheo.network/': 'https://ropsten.infura.io/'
        }
        self.infura_url = self.infura_dict[api_url]
        self.blockchain_amount = {
            'eth': partial(to_wei, unit='ether'),
            'neo': to_neo_asset_amount
        }
        self.sign_create_cancellation_function = {
            'eth': sign_create_cancellation_eth,
            'neo': sign_create_cancellation_neo
        }
        self.sign_execute_cancellation_function = {
            'eth': sign_execute_cancellation_eth,
            'neo': sign_execute_cancellation_neo
        }
        self.sign_create_deposit_function = {
            'eth': sign_create_deposit_eth,
            'neo': sign_create_deposit_neo
        }
        self.sign_execute_deposit_function = {
            'eth': partial(sign_execute_deposit_eth, infura_url=self.infura_url),
            'neo': sign_execute_deposit_neo
        }
        self.sign_create_order_function = {
            'eth': sign_create_order_eth,
            'neo': sign_create_order_neo
        }
        self.sign_execute_order_function = {
            'eth': sign_execute_order_eth,
            'neo': sign_execute_order_neo
        }
        self.sign_create_withdrawal_function = {
            'eth': sign_create_withdrawal_eth,
            'neo': sign_create_withdrawal_neo
        }
        self.sign_execute_withdrawal_function = {
            'eth': sign_execute_withdrawal_eth,
            'neo': sign_execute_withdrawal_neo
        }

    def cancel_order(self, order_id, private_key):
        """
        This function is a wrapper function around the create and execute cancellation functions to help make this
        processes simpler for the end user by combining these requests in 1 step.
        Execution of this function is as follows::

            cancel_order(order_id=order['id'], private_key=kp)
            cancel_order(order_id=order['id'], private_key=eth_private_key)

        The expected return result for this function is the same as the execute_cancellation function::

            {
                'id': 'b8e617d5-f5ed-4600-b8f2-7d370d837750',
                'blockchain': 'neo',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'side': 'buy',
                'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'offer_amount': '2000000',
                'want_amount': '10000000000',
                'transfer_amount': '0',
                'priority_gas_amount': '0',
                'use_native_token': True,
                'native_fee_transfer_amount': 0,
                'deposit_txn': None,
                'created_at': '2018-08-05T11:16:47.021Z',
                'status': 'processed',
                'fills': [],
                'makes': [
                    {
                        'id': '6b9f40de-f9bb-46b6-9434-d281f8c06b74',
                        'offer_hash': '6830d82dbdda566ab32e9a8d9d9d94d3297f67c10374d69bb35d6c5a86bd3e92',
                        'available_amount': '0',
                        'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                        'offer_amount': '2000000',
                        'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                        'want_amount': '10000000000',
                        'filled_amount': '0.0',
                        'txn': None,
                        'cancel_txn': None,
                        'price': '0.0002',
                        'status': 'cancelling',
                        'created_at': '2018-08-05T11:16:47.036Z',
                        'transaction_hash': 'e5b08c4a55c7494f1ec7dd93ac2bb2b4e84e77dec9e00e91be1d520cb818c415',
                        'trades': []
                    }
                ]
            }

        :param order_id: The order ID of the open transaction on the order book that you want to cancel.
        :type order_id: str
        :param private_key: The KeyPair that will be used to sign the transaction sent to the blockchain.
        :type private_key: KeyPair
        :return: Dictionary of the transaction details and state after sending the signed transaction to the blockchain.
        """
        create_cancellation = self.create_cancellation(order_id=order_id, private_key=private_key)
        return self.execute_cancellation(cancellation_params=create_cancellation, private_key=private_key)

    def create_cancellation(self, order_id, private_key):
        """
        Function to create a cancellation request for the order ID from the open orders on the order book.
        Execution of this function is as follows::

            create_cancellation(order_id=order['id'], private_key=kp)

        The expected return result for this function is as follows::

            {
                'id': '6b9f40de-f9bb-46b6-9434-d281f8c06b74',
                'transaction': {
                    'hash': '50d99ebd7e57dbdceb7edc2014da5f446c8f44cc0a0b6d9c762a29e8a74bb051',
                    'sha256': '509edb9888fa675988fa71a27600b2655e63fe979424f13f5c958897b2e99ed8',
                    'type': 209,
                    'version': 1,
                    'attributes': [
                        {
                            'usage': 32,
                            'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                        }
                    ],
                    'inputs': [
                        {
                            'prevHash': '9eaca1adcbbc0669a936576cb9ad03c11c99c356347aae3037ce1f0e4d330d85',
                            'prevIndex': 0
                        }, {
                            'prevHash': 'c858e4d2af1e1525fa974fb2b1678caca1f81a5056513f922789594939ff713d',
                            'prevIndex': 37
                        }
                    ],
                    'outputs': [
                        {
                            'assetId': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
                            'scriptHash': 'e707714512577b42f9a011f8b870625429f93573',
                            'value': 1e-08
                        }
                    ],
                    'scripts': [],
                    'script': '....',
                    'gas': 0
                },
                'script_params': {
                    'scriptHash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                    'operation': 'cancelOffer',
                    'args': [
                        '923ebd865a6c5db39bd67403c1677f29d3949d9d8d9a2eb36a56dabd2dd83068'
                    ]
                }
            }

        :param order_id: The order ID of the open transaction on the order book that you want to cancel.
        :type order_id: str
        :param private_key: The KeyPair that will be used to sign the transaction sent to the blockchain.
        :type private_key: KeyPair
        :return: Dictionary that contains the cancellation request along with blockchain transaction information.
        """
        cancellation_params = {
            "order_id": order_id,
            "timestamp": get_epoch_milliseconds()
        }
        api_params = self.sign_create_cancellation_function[self.blockchain](cancellation_params, private_key)
        return self.request.post(path='/cancellations', json_data=api_params)

    def execute_cancellation(self, cancellation_params, private_key):
        """
        This function executes the order created before it and signs the transaction to be submitted to the blockchain.
        Execution of this function is as follows::

            execute_cancellation(cancellation_params=create_cancellation, private_key=kp)

        The expected return result for this function is as follows::

            {
                'id': 'b8e617d5-f5ed-4600-b8f2-7d370d837750',
                'blockchain': 'neo',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'side': 'buy',
                'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'offer_amount': '2000000',
                'want_amount': '10000000000',
                'transfer_amount': '0',
                'priority_gas_amount': '0',
                'use_native_token': True,
                'native_fee_transfer_amount': 0,
                'deposit_txn': None,
                'created_at': '2018-08-05T11:16:47.021Z',
                'status': 'processed',
                'fills': [],
                'makes': [
                    {
                        'id': '6b9f40de-f9bb-46b6-9434-d281f8c06b74',
                        'offer_hash': '6830d82dbdda566ab32e9a8d9d9d94d3297f67c10374d69bb35d6c5a86bd3e92',
                        'available_amount': '0',
                        'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                        'offer_amount': '2000000',
                        'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                        'want_amount': '10000000000',
                        'filled_amount': '0.0',
                        'txn': None,
                        'cancel_txn': None,
                        'price': '0.0002',
                        'status': 'cancelling',
                        'created_at': '2018-08-05T11:16:47.036Z',
                        'transaction_hash': 'e5b08c4a55c7494f1ec7dd93ac2bb2b4e84e77dec9e00e91be1d520cb818c415',
                        'trades': []
                    }
                ]
            }

        :param cancellation_params: Parameters generated from the Switcheo API to cancel an order on the order book.
        :type cancellation_params: dict
        :param private_key: The KeyPair that will be used to sign the transaction sent to the blockchain.
        :type private_key: KeyPair
        :return: Dictionary of the transaction details and state after sending the signed transaction to the blockchain.
        """
        cancellation_id = cancellation_params['id']
        api_params = self.sign_execute_cancellation_function[self.blockchain](cancellation_params, private_key)
        return self.request.post(path='/cancellations/{}/broadcast'.format(cancellation_id), json_data=api_params)

    def deposit(self, asset, amount, private_key):
        """
        This function is a wrapper function around the create and execute deposit functions to help make this
        processes simpler for the end user by combining these requests in 1 step.
        Execution of this function is as follows::

            deposit(asset="SWTH", amount=1.1, private_key=KeyPair)
            deposit(asset="SWTH", amount=1.1, private_key=eth_private_key)

        The expected return result for this function is the same as the execute_deposit function::

            {
                'result': 'ok'
            }

        :param asset: Symbol or Script Hash of asset ID from the available products.
        :type asset: str
        :param amount: The amount of coins/tokens to be deposited.
        :type amount: float
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :return: Dictionary with the result status of the deposit attempt.
        """
        create_deposit = self.create_deposit(asset=asset, amount=amount, private_key=private_key)
        return self.execute_deposit(deposit_params=create_deposit, private_key=private_key)

    def create_deposit(self, asset, amount, private_key):
        """
        Function to create a deposit request by generating a transaction request from the Switcheo API.
        Execution of this function is as follows::

            create_deposit(asset="SWTH", amount=1.1, private_key=KeyPair)
            create_deposit(asset="ETH", amount=1.1, private_key=eth_private_key)

        The expected return result for this function is as follows::

            {
                'id': '768e2079-1504-4dad-b688-7e1e99ec0a24',
                'transaction': {
                    'hash': '72b74c96b9174e9b9e1b216f7e8f21a6475e6541876a62614df7c1998c6e8376',
                    'sha256': '2109cbb5eea67a06f5dd8663e10fcd1128e28df5721a25d993e05fe2097c34f3',
                    'type': 209,
                    'version': 1,
                    'attributes': [
                        {
                            'usage': 32,
                            'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                        }
                    ],
                    'inputs': [
                        {
                            'prevHash': 'f09b3b697c580d1730cd360da5e1f0beeae00827eb2f0055cbc85a5a4dadd8ea',
                            'prevIndex': 0
                        }, {
                            'prevHash': 'c858e4d2af1e1525fa974fb2b1678caca1f81a5056513f922789594939ff713d',
                            'prevIndex': 31
                        }
                    ],
                    'outputs': [
                        {
                            'assetId': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
                            'scriptHash': 'e707714512577b42f9a011f8b870625429f93573',
                            'value': 1e-08
                        }
                    ],
                    'scripts': [],
                    'script': '....',
                    'gas': 0
                },
                'script_params': {
                    'scriptHash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                    'operation': 'deposit',
                    'args': [
                        '592c8a46a0d06c600f06c994d1f25e7283b8a2fe',
                        '32e125258b7db0a0dffde5bd03b2b859253538ab',
                        100000000
                    ]
                }
            }

        :param asset: Symbol or Script Hash of asset ID from the available products.
        :type asset: str
        :param amount: The amount of coins/tokens to be deposited.
        :type amount: float
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :return: Dictionary response of signed deposit request that is ready to be executed on the specified blockchain.
        """
        signable_params = {
            'blockchain': self.blockchain,
            'asset_id': asset,
            'amount': str(self.blockchain_amount[self.blockchain](amount)),
            'timestamp': get_epoch_milliseconds(),
            'contract_hash': self.contract_hash
        }
        api_params = self.sign_create_deposit_function[self.blockchain](signable_params, private_key)
        return self.request.post(path='/deposits', json_data=api_params)

    def execute_deposit(self, deposit_params, private_key):
        """
        Function to execute the deposit request by signing the transaction generated by the create deposit function.
        Execution of this function is as follows::

            execute_deposit(deposit_params=create_deposit, private_key=KeyPair)
            execute_deposit(deposit_params=create_deposit, private_key=eth_private_key)

        The expected return result for this function is as follows::

            {
                'result': 'ok'
            }

        :param deposit_params: Parameters from the API to be signed and deposited to the Switcheo Smart Contract.
        :type deposit_params: dict
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :return: Dictionary with the result status of the deposit attempt.
        """
        deposit_id = deposit_params['id']
        api_params = self.sign_execute_deposit_function[self.blockchain](deposit_params, private_key)
        return self.request.post(path='/deposits/{}/broadcast'.format(deposit_id), json_data=api_params)

    def order(self, pair, side, price, quantity, private_key, use_native_token=True, order_type="limit"):
        """
        This function is a wrapper function around the create and execute order functions to help make this processes
        simpler for the end user by combining these requests in 1 step.
        Execution of this function is as follows::

            order(pair="SWTH_NEO", side="buy",
                  price=0.0002, quantity=100, private_key=kp,
                  use_native_token=True, order_type="limit")

        The expected return result for this function is the same as the execute_order function::

            {
                'id': '4e6a59fd-d750-4332-aaf0-f2babfa8ad67',
                'blockchain': 'neo',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'side': 'buy',
                'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'offer_amount': '2000000',
                'want_amount': '10000000000',
                'transfer_amount': '0',
                'priority_gas_amount': '0',
                'use_native_token': True,
                'native_fee_transfer_amount': 0,
                'deposit_txn': None,
                'created_at': '2018-08-05T10:38:37.714Z',
                'status': 'processed',
                'fills': [],
                'makes': [
                    {
                        'id': 'e30a7fdf-779c-4623-8f92-8a961450d843',
                        'offer_hash': 'b45ddfb97ade5e0363d9e707dac9ad1c530448db263e86494225a0025006f968',
                        'available_amount': '2000000',
                        'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                        'offer_amount': '2000000',
                        'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                        'want_amount': '10000000000',
                        'filled_amount': '0.0',
                        'txn': None,
                        'cancel_txn': None,
                        'price': '0.0002',
                        'status': 'confirming',
                        'created_at': '2018-08-05T10:38:37.731Z',
                        'transaction_hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                        'trades': []
                    }
                ]
            }

        :param pair: The trading pair this order is being submitted for.
        :type pair: str
        :param side: The side of the trade being submitted i.e. buy or sell
        :type side: str
        :param price: The price target for this trade.
        :type price: float
        :param quantity: The amount of the asset being exchanged in the trade.
        :type quantity: float
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :param use_native_token: Flag to indicate whether or not to pay fees with the Switcheo native token.
        :type use_native_token: bool
        :param order_type: The type of order being submitted, currently this can only be a limit order.
        :type order_type: str
        :return: Dictionary of the transaction on the order book.
        """
        create_order = self.create_order(private_key=private_key, pair=pair, side=side, price=price,
                                         quantity=quantity, use_native_token=use_native_token,
                                         order_type=order_type)
        return self.execute_order(order_params=create_order, private_key=private_key)

    def create_order(self, pair, side, price, quantity, private_key, use_native_token=True, order_type="limit",
                     otc_address=None):
        """
        Function to create an order for the trade pair and details requested.
        Execution of this function is as follows::

            create_order(pair="SWTH_NEO", side="buy", price=0.0002, quantity=100, private_key=kp,
                         use_native_token=True, order_type="limit")

        The expected return result for this function is as follows::

            {
                'id': '4e6a59fd-d750-4332-aaf0-f2babfa8ad67',
                'blockchain': 'neo',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'side': 'buy',
                'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'offer_amount': '2000000',
                'want_amount': '10000000000',
                'transfer_amount': '0',
                'priority_gas_amount': '0',
                'use_native_token': True,
                'native_fee_transfer_amount': 0,
                'deposit_txn': None,
                'created_at': '2018-08-05T10:38:37.714Z',
                'status': 'pending',
                'fills': [],
                'makes': [
                    {
                        'id': 'e30a7fdf-779c-4623-8f92-8a961450d843',
                        'offer_hash': None,
                        'available_amount': None,
                        'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                        'offer_amount': '2000000',
                        'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                        'want_amount': '10000000000',
                        'filled_amount': None,
                        'txn': {
                            'offerHash': 'b45ddfb97ade5e0363d9e707dac9ad1c530448db263e86494225a0025006f968',
                            'hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                            'sha256': 'f0b70640627947584a2976edeb055a124ae85594db76453532b893c05618e6ca',
                            'invoke': {
                                'scriptHash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                                'operation': 'makeOffer',
                                'args': [
                                    '592c8a46a0d06c600f06c994d1f25e7283b8a2fe',
                                    '9b7cffdaa674beae0f930ebe6085af9093e5fe56b34a5c220ccdcf6efc336fc5',
                                    2000000,
                                    '32e125258b7db0a0dffde5bd03b2b859253538ab',
                                    10000000000,
                                    '65333061376664662d373739632d343632332d386639322d386139363134353064383433'
                                ]
                            },
                            'type': 209,
                            'version': 1,
                            'attributes': [
                                {
                                    'usage': 32,
                                    'data': '592c8a46a0d06c600f06c994d1f25e7283b8a2fe'
                                }
                            ],
                            'inputs': [
                                {
                                    'prevHash': '0fcfd792a9d20a7795255d1d3d3927f5968b9953e80d16ffd222656edf8fedbc',
                                    'prevIndex': 0
                                }, {
                                    'prevHash': 'c858e4d2af1e1525fa974fb2b1678caca1f81a5056513f922789594939ff713d',
                                    'prevIndex': 35
                                }
                            ],
                            'outputs': [
                                {
                                    'assetId': '602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
                                    'scriptHash': 'e707714512577b42f9a011f8b870625429f93573',
                                    'value': 1e-08
                                }
                            ],
                            'scripts': [],
                            'script': '....',
                            'gas': 0
                        },
                        'cancel_txn': None,
                        'price': '0.0002',
                        'status': 'pending',
                        'created_at': '2018-08-05T10:38:37.731Z',
                        'transaction_hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                        'trades': []
                    }
                ]
            }

        :param pair: The trading pair this order is being submitted for.
        :type pair: str
        :param side: The side of the trade being submitted i.e. buy or sell
        :type side: str
        :param price: The price target for this trade.
        :type price: float
        :param quantity: The amount of the asset being exchanged in the trade.
        :type quantity: float
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :param use_native_token: Flag to indicate whether or not to pay fees with the Switcheo native token.
        :type use_native_token: bool
        :param order_type: The type of order being submitted, currently this can only be a limit order.
        :type order_type: str
        :param otc_address: The address to trade with for Over the Counter exchanges.
        :type otc_address: str
        :return: Dictionary of order details to specify which parts of the trade will be filled (taker) or open (maker)
        """
        if side.lower() not in ["buy", "sell"]:
            raise ValueError("Allowed trade types are buy or sell, you entered {}".format(side.lower()))
        if order_type.lower() not in ["limit", "market", "otc"]:
            raise ValueError("Allowed order type is limit, you entered {}".format(order_type.lower()))
        if order_type.lower() == "otc" and otc_address is None:
            raise ValueError("OTC Address is required when trade type is otc (over the counter).")
        order_params = {
            "blockchain": self.blockchain,
            "pair": pair,
            "side": side,
            "price": '{:.8f}'.format(price) if order_type.lower() != "market" else None,
            "quantity": str(self.blockchain_amount[self.blockchain](quantity)),
            "use_native_tokens": use_native_token,
            "order_type": order_type,
            "timestamp": get_epoch_milliseconds(),
            "contract_hash": self.contract_hash
        }
        api_params = self.sign_create_order_function[self.blockchain](order_params, private_key)
        return self.request.post(path='/orders', json_data=api_params)

    def execute_order(self, order_params, private_key):
        """
        This function executes the order created before it and signs the transaction to be submitted to the blockchain.
        Execution of this function is as follows::

            execute_order(order_params=create_order, private_key=kp)

        The expected return result for this function is the same as the execute_order function::

            {
                'id': '4e6a59fd-d750-4332-aaf0-f2babfa8ad67',
                'blockchain': 'neo',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82',
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'side': 'buy',
                'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'offer_amount': '2000000',
                'want_amount': '10000000000',
                'transfer_amount': '0',
                'priority_gas_amount': '0',
                'use_native_token': True,
                'native_fee_transfer_amount': 0,
                'deposit_txn': None,
                'created_at': '2018-08-05T10:38:37.714Z',
                'status': 'processed',
                'fills': [],
                'makes': [
                    {
                        'id': 'e30a7fdf-779c-4623-8f92-8a961450d843',
                        'offer_hash': 'b45ddfb97ade5e0363d9e707dac9ad1c530448db263e86494225a0025006f968',
                        'available_amount': '2000000',
                        'offer_asset_id': 'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
                        'offer_amount': '2000000',
                        'want_asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                        'want_amount': '10000000000',
                        'filled_amount': '0.0',
                        'txn': None,
                        'cancel_txn': None,
                        'price': '0.0002',
                        'status': 'confirming',
                        'created_at': '2018-08-05T10:38:37.731Z',
                        'transaction_hash': '5c4cb1e73b9f2e608b6e768e0654649a4d15e08a7fe63fc536c454fa563a2f0f',
                        'trades': []
                    }
                ]
            }

        :param order_params: Dictionary generated from the create order function.
        :type order_params: dict
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :return: Dictionary of the transaction on the order book.
        """
        order_id = order_params['id']
        api_params = self.sign_execute_order_function[self.blockchain](order_params, private_key)
        return self.request.post(path='/orders/{}/broadcast'.format(order_id), json_data=api_params)

    def withdrawal(self, asset, amount, private_key):
        """
        This function is a wrapper function around the create and execute withdrawal functions to help make this
        processes simpler for the end user by combining these requests in 1 step.
        Execution of this function is as follows::

            withdrawal(asset="SWTH", amount=1.1, private_key=kp))

        The expected return result for this function is the same as the execute_withdrawal function::

            {
                'event_type': 'withdrawal',
                'amount': -100000,
                'asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'status': 'confirming',
                'id': '96e5f797-435b-40ab-9085-4e95c6749218',
                'blockchain': 'neo',
                'reason_code': 9,
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'transaction_hash': None,
                'created_at': '2018-08-05T10:03:58.885Z',
                'updated_at': '2018-08-05T10:03:59.828Z',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82'
            }

        :param asset: Script Hash of asset ID from the available products.
        :type asset: str
        :param amount: The amount of coins/tokens to be withdrawn.
        :type amount: float
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :return: Dictionary with the status of the withdrawal request and blockchain details.
        """
        create_withdrawal = self.create_withdrawal(asset=asset, amount=amount, private_key=private_key)
        return self.execute_withdrawal(withdrawal_params=create_withdrawal, private_key=private_key)

    def create_withdrawal(self, asset, amount, private_key):
        """
        Function to create a withdrawal request by generating a withdrawal ID request from the Switcheo API.
        Execution of this function is as follows::

            create_withdrawal(asset="SWTH", amount=1.1, private_key=kp)

        The expected return result for this function is as follows::

            {
                'id': 'a5a4d396-fa9f-4191-bf50-39a3d06d5e0d'
            }

        :param asset:  Script Hash of asset ID from the available products.
        :type asset: str
        :param amount: The amount of coins/tokens to be withdrawn.
        :type amount: float
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :return: Dictionary with the withdrawal ID generated by the Switcheo API.
        """
        signable_params = {
            'blockchain': self.blockchain,
            'asset_id': asset,
            'amount': str(self.blockchain_amount[self.blockchain](amount)),
            'timestamp': get_epoch_milliseconds(),
            'contract_hash': self.contract_hash
        }
        api_params = self.sign_create_withdrawal_function[self.blockchain](signable_params, private_key)
        return self.request.post(path='/withdrawals', json_data=api_params)

    def execute_withdrawal(self, withdrawal_params, private_key):
        """
        This function is to sign the message generated from the create withdrawal function and submit it to the
        blockchain for transfer from the smart contract to the owners address.
        Execution of this function is as follows::

            execute_withdrawal(withdrawal_params=create_withdrawal, private_key=kp)

        The expected return result for this function is as follows::

            {
                'event_type': 'withdrawal',
                'amount': -100000,
                'asset_id': 'ab38352559b8b203bde5fddfa0b07d8b2525e132',
                'status': 'confirming',
                'id': '96e5f797-435b-40ab-9085-4e95c6749218',
                'blockchain': 'neo',
                'reason_code': 9,
                'address': 'fea2b883725ef2d194c9060f606cd0a0468a2c59',
                'transaction_hash': None,
                'created_at': '2018-08-05T10:03:58.885Z',
                'updated_at': '2018-08-05T10:03:59.828Z',
                'contract_hash': 'a195c1549e7da61b8da315765a790ac7e7633b82'
            }

        :param withdrawal_params: Dictionary from the create withdrawal function to sign and submit to the blockchain.
        :type withdrawal_params: dict
        :param private_key: The Private Key (ETH) or KeyPair (NEO) for the wallet being used to sign deposit message.
        :type private_key: KeyPair or str
        :return: Dictionary with the status of the withdrawal request and blockchain transaction details.
        """
        withdrawal_id = withdrawal_params['id']
        api_params = self.sign_execute_withdrawal_function[self.blockchain](withdrawal_params, private_key)
        return self.request.post(path='/withdrawals/{}/broadcast'.format(withdrawal_id), json_data=api_params)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
try:  # pip version >= 10.0
    from pip._internal.req import parse_requirements
    from pip._internal.download import PipSession
except ImportError:  # pip version < 10.0
    from pip.req import parse_requirements
    from pip.download import PipSession

install_reqs = parse_requirements('requirements.txt', session=PipSession())
reqs = [str(ir.req) for ir in install_reqs]

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='switcheo',
    python_requires='>=3.6',
    version='0.3.4',
    author='Keith Smith',
    author_email='keith.scotts@gmail.com',
    license='MIT License',
    url='https://github.com/KeithSSmith/switcheo-python',
    packages=find_packages(),
    include_package_data=True,
    install_requires=reqs,
    description='Python Client to interact with the Switcheo Exchange API',
    long_description=long_description,
    keywords=['switcheo', 'switcheo-api', 'trade', 'ethereum', 'neo', 'ETH', 'NEO',
              'QTUM', 'client', 'api', 'wrapper', 'exchange', 'dex', 'crypto', 'currency', 'trading', 'trading-api'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import binascii
from eth_account.account import Account


def sign_txn_array(messages, private_key):
    message_dict = {}
    for message in messages:
        signed_message =\
            binascii.hexlify(Account.signHash(message['txn']['sha256'], private_key=private_key)['signature']).decode()
        message_dict[message['id']] = '0x' + signed_message
    return message_dict



# -*- coding:utf-8 -*-
"""
Description:
    Switcheo Client is designed to standardize interactions with the Python Client.
    It can access the Public and Authenticated Clients and is designed to be more user friendly than the
    forward facing REST API's.
    Ideally, more simplified/advanced trading functions will be built here (trailing stop, all or none, etc)
Usage:
    from switcheo.switcheo_client import SwitcheoClient
"""

from switcheo.utils import current_contract_hash
from switcheo.authenticated_client import AuthenticatedClient
from switcheo.public_client import PublicClient
from switcheo.neo.utils import neo_get_scripthash_from_address

contract_version_dict = {
    "eth": 'V1',
    "neo": 'V3'
}

network_dict = {
    "neo": "neo",
    "NEO": "neo",
    "eth": "eth",
    "ETH": "eth",
    "ethereum": "eth",
    "Ethereum": "eth"
}

url_dict = {
    "main": 'https://api.switcheo.network/',
    "test": 'https://test-api.switcheo.network/'
}


class SwitcheoClient(AuthenticatedClient, PublicClient):

    def __init__(self,
                 switcheo_network="test",
                 blockchain_network="neo",
                 private_key=None):
        self.api_url = url_dict[switcheo_network]
        self.blockchain = network_dict[blockchain_network]
        self.contract_version = contract_version_dict[self.blockchain]
        super().__init__(blockchain=self.blockchain,
                         contract_version=self.contract_version,
                         api_url=self.api_url)
        self.private_key = private_key

    def order_history(self, address, pair=None):
        return self.get_orders(neo_get_scripthash_from_address(address=address), pair=pair)

    def balance_current_contract(self, *addresses):
        address_list = []
        contract_dict = {}
        for address in addresses:
            address_list.append(neo_get_scripthash_from_address(address=address))
        current_contract = current_contract_hash(self.contracts)
        for chain in current_contract.keys():
            contract_dict[chain] =\
                self.get_balance(addresses=address_list, contracts=current_contract[chain])
        return contract_dict

    def balance_by_contract(self, *addresses):
        address_list = []
        contract_dict = {}
        for address in addresses:
            address_list.append(neo_get_scripthash_from_address(address=address))
        contracts = self.get_contracts()
        for blockchain in contracts:
            contract_dict[blockchain] = {}
            for key in contracts[blockchain]:
                contract_dict[blockchain][key] =\
                    self.get_balance(addresses=address_list, contracts=contracts[blockchain][key])
        return contract_dict

    def balance_by_address_by_contract(self, *addresses):
        contract_dict = {}
        for address in addresses:
            contract_dict[address] = self.balance_by_contract(address)
        return contract_dict

    def limit_buy(self, price, quantity, pair, use_native_token=True):
        """

            limit_buy(price=0.0002, quantity=1000, pair='SWTH_NEO')
            limit_buy(price=0.0000001, quantity=1000000, pair='JRC_ETH')

        :param price:
        :param quantity:
        :param pair:
        :param use_native_token:
        :return:
        """
        if 'ETH' in pair:
            use_native_token = False
        return self.order(order_type="limit",
                          side="buy",
                          pair=pair,
                          price=price,
                          quantity=quantity,
                          private_key=self.private_key,
                          use_native_token=use_native_token)

    def limit_sell(self, price, quantity, pair, use_native_token=True):
        """

            limit_sell(price=0.0006, quantity=500, pair='SWTH_NEO')
            limit_sell(price=0.000001, quantity=100000, pair='JRC_ETH')

        :param price:
        :param quantity:
        :param pair:
        :param use_native_token:
        :return:
        """
        if 'ETH' in pair:
            use_native_token = False
        return self.order(order_type="limit",
                          side="sell",
                          pair=pair,
                          price=price,
                          quantity=quantity,
                          private_key=self.private_key,
                          use_native_token=use_native_token)

    def market_buy(self, quantity, pair, use_native_token=True):
        """

            market_buy(quantity=100, pair='SWTH_NEO')
            market_buy(quantity=100000, pair='JRC_ETH')

        :param quantity:
        :param pair:
        :param use_native_token:
        :return:
        """
        if 'ETH' in pair:
            use_native_token = False
        return self.order(order_type="market",
                          side="buy",
                          pair=pair,
                          price=0,
                          quantity=quantity,
                          private_key=self.private_key,
                          use_native_token=use_native_token)

    def market_sell(self, quantity, pair, use_native_token=True):
        """

            market_sell(quantity=100, pair='SWTH_NEO')
            market_sell(quantity=100000, pair='JRC_ETH')

        :param quantity:
        :param pair:
        :param use_native_token:
        :return:
        """
        if 'ETH' in pair:
            use_native_token = False
        return self.order(order_type="market",
                          side="sell",
                          pair=pair,
                          price=0,
                          quantity=quantity,
                          private_key=self.private_key,
                          use_native_token=use_native_token)



#
# switcheo/neo/transactions.py
# Keith Smith
#
# For testnet requests to the Switcheo exchange


from switcheo.utils import reverse_hex, num2hexstring, num2varint
from switcheo.Fixed8 import SwitcheoFixed8, num2fixed8


max_transaction_attribute_size = 65535


def serialize_transaction(transaction, signed=True):
    serialized_txn = ''
    serialized_txn += str(num2hexstring(transaction['type']))
    serialized_txn += str(num2hexstring(transaction['version']))
    serialized_txn += str(serialize_exclusive[transaction['type']](transaction))
    serialized_txn += str(num2varint(len(transaction['attributes'])))
    for attribute in transaction['attributes']:
        serialized_txn += serialize_transaction_attribute(attribute)
    serialized_txn += str(num2varint(len(transaction['inputs'])))
    for txn_input in transaction['inputs']:
        serialized_txn += serialize_transaction_input(txn_input)
    serialized_txn += str(num2varint(len(transaction['outputs'])))
    for txn_output in transaction['outputs']:
        serialized_txn += serialize_transaction_output(txn_output)
    if signed and transaction['scripts'] and len(transaction['scripts']) > 0:
        serialized_txn += str(num2varint(len(transaction['scripts'])))
        for script in transaction['scripts']:
            serialized_txn += serialize_witness(script)
    return serialized_txn


def serialize_transaction_attribute(attr):
    if len(attr['data']) > max_transaction_attribute_size:
        raise ValueError('Transaction attribute data is larger than the Maximum allowed attribute size.')
    out = num2hexstring(attr['usage'])
    if attr['usage'] == 0x81:
        out += num2hexstring(len(attr['data']) // 2)
    elif attr['usage'] == 0x90 or attr['usage'] >= 0xf0:
        out += num2varint(len(attr['data']) // 2)
    if attr['usage'] == 0x02 or attr['usage'] == 0x03:
        out += attr['data'][2:64]
    else:
        out += attr['data']
    return out


def serialize_transaction_input(txn_input):
    return reverse_hex(txn_input['prevHash']) + reverse_hex(num2hexstring(txn_input['prevIndex'], 2))


def serialize_transaction_output(txn_output):
    value = SwitcheoFixed8(float(txn_output['value'])).toReverseHex()
    return reverse_hex(txn_output['assetId']) + value + reverse_hex(txn_output['scriptHash'])


def serialize_witness(witness):
    invocation_len = num2varint(len(witness['invocationScript']) // 2)
    verification_len = num2varint(len(witness['verificationScript']) // 2)
    return invocation_len + witness['invocationScript'] + verification_len + witness['verificationScript']


# Switcheo does not allow for GAS claims so this function should not be used.
def serialize_claim_exclusive(transaction):
    if transaction['type'] != 0x02:
        raise ValueError(
            'The transaction type {} does not match the claim exclusive method.'.format(transaction['type']))
    out = num2varint(len(transaction['claims']))
    for claim in transaction['claims']:
        out += serialize_transaction_input(claim)
    return out


def serialize_contract_exclusive(transaction):
    if transaction['type'] != 0x80:
        raise ValueError(
            'The transaction type {} does not match the contract exclusive method.'.format(transaction['type']))
    return ''


def serialize_invocation_exclusive(transaction):
    if transaction['type'] != 0xd1:
        raise ValueError(
            'The transaction type {} does not match the invocation exclusive method.'.format(transaction['type']))
    out = num2varint(int(len(transaction['script'])/2))
    out += transaction['script']
    if transaction['version'] >= 1:
        out += num2fixed8(transaction['gas'])
    return out


serialize_exclusive = {
  2: serialize_claim_exclusive,
  128: serialize_contract_exclusive,
  209: serialize_invocation_exclusive
}



#
# switcheo/utils.py
# Keith Smith
#
# For testnet requests to the Switcheo exchange

import json
import requests
import time


def get_epoch_milliseconds():
    return int(round(time.time() * 1000))


def stringify_message(message):
    """Return a JSON message that is alphabetically sorted by the key name

    Args:
        message
    """
    return json.dumps(message, sort_keys=True, separators=(',', ':'))


def reverse_hex(message):
    return "".join([message[x:x + 2] for x in range(0, len(message), 2)][::-1])


def num2hexstring(number, size=1, little_endian=False):
    """
    Converts a number to a big endian hexstring of a suitable size, optionally little endian
    :param {number} number
    :param {number} size - The required size in hex chars, eg 2 for Uint8, 4 for Uint16. Defaults to 2.
    :param {boolean} little_endian - Encode the hex in little endian form
    :return {string}
    """
    # if (type(number) != = 'number') throw new Error('num must be numeric')
    # if (num < 0) throw new RangeError('num is unsigned (>= 0)')
    # if (size % 1 !== 0) throw new Error('size must be a whole integer')
    # if (!Number.isSafeInteger(num)) throw new RangeError(`num (${num}) must be a safe integer`)
    size = size * 2
    hexstring = hex(number)[2:]
    if len(hexstring) % size != 0:
        hexstring = ('0' * size + hexstring)[len(hexstring):]
    if little_endian:
        hexstring = reverse_hex(hexstring)
    return hexstring


def num2varint(num):
    """
    Converts a number to a variable length Int. Used for array length header

    :param: {number} num - The number
    :return: {string} hexstring of the variable Int.
    """
    # if (typeof num !== 'number') throw new Error('VarInt must be numeric')
    # if (num < 0) throw new RangeError('VarInts are unsigned (> 0)')
    # if (!Number.isSafeInteger(num)) throw new RangeError('VarInt must be a safe integer')
    if num < 0xfd:
        return num2hexstring(num)
    elif num <= 0xffff:
        # uint16
        return 'fd' + num2hexstring(number=num, size=2, little_endian=True)
    elif num <= 0xffffffff:
        # uint32
        return 'fe' + num2hexstring(number=num, size=4, little_endian=True)
    else:
        # uint64
        return 'ff' + num2hexstring(number=num, size=8, little_endian=True)


def current_contract_hash(contracts):
    contract_dict = {}
    for chain in contracts:
        max_key = 1
        for key in contracts[chain].keys():
            if float(key[1:].replace('_', '.')) > max_key:
                max_key = float(key[1:].replace('_', '.'))
        max_key_str = 'V' + str(max_key).replace('.', '_').replace('_0', '')
        contract_dict[chain] = contracts[chain][max_key_str]
    return contract_dict



class SwitcheoApiException(Exception):

    def __init__(self, error_code, error_message, error):
        super(SwitcheoApiException, self).__init__(error_message)
        self.error_code = error_code
        self.error = error
        

class Request(object):

    def __init__(self, api_url='https://test-api.switcheo.network/', api_version="/v2", timeout=30):
        self.base_url = api_url.rstrip('/')
        self.url = self.base_url + api_version
        self.timeout = timeout

    def get(self, path, params=None):
        """Perform GET request"""
        r = requests.get(url=self.url + path, params=params, timeout=self.timeout)
        r.raise_for_status()
        return r.json()

    def post(self, path, data=None, json_data=None, params=None):
        """Perform POST request"""
        r = requests.post(url=self.url + path, data=data, json=json_data, params=params, timeout=self.timeout)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError:
            raise SwitcheoApiException(r.json().get('error_code'), r.json().get('error_message'), r.json().get('error'))
        return r.json()

    def status(self):
        r = requests.get(url=self.base_url)
        r.raise_for_status()
        return r.json()
